"""SQLAlchemy pipeline optimizer.

Analyzes pipelines and optimizes them to execute as SQL queries when possible.
"""

from typing import Any

from haolib.pipelines.base import Operation, Pipeline, TargetBoundOperation, TargetSwitch
from haolib.pipelines.context import PipelineContext

# Import operations lazily to avoid circular import
# Operations are imported in methods that use them
from haolib.pipelines.optimizer import PipelineAnalysis
from haolib.storages.data_types.registry import DataTypeRegistry
from haolib.storages.indexes.sqlalchemy import IndexHandler
from haolib.storages.operations.optimizer.predicate_analyzer import PredicateAnalyzer
from haolib.storages.operations.optimizer.query_builder import QueryBuilder


class SQLAlchemyPipelineOptimizer:
    """Optimizer for SQLAlchemy pipelines.

    Analyzes pipelines and optimizes them to execute as SQL queries when possible.
    Can convert operations like filter, map, reduce into SQL WHERE, SELECT, and
    aggregation clauses.

    Example:
        ```python
        optimizer = SQLAlchemyPipelineOptimizer(registry)
        analysis = optimizer.analyze(pipeline)

        if analysis.execution_plan == "storage":
            # Entire pipeline can be executed in SQL
            result = await storage.execute(analysis.optimized_operation)
        elif analysis.execution_plan == "hybrid":
            # Part in SQL, part in Python
            sql_result = await storage.execute(analysis.optimized_operation)
            # Then execute remaining operations in Python
        else:
            # All in Python
            result = await storage.execute(pipeline)
        ```

    """

    def __init__(self, registry: DataTypeRegistry) -> None:
        """Initialize the optimizer.

        Args:
            registry: Data type registry for converting user types to storage models.

        """
        self._registry = registry
        self._index_handler = IndexHandler(registry)
        self._predicate_analyzer = PredicateAnalyzer()
        self._query_builder = QueryBuilder(registry)

    def analyze(self, pipeline: Any, pipeline_context: PipelineContext | None = None) -> PipelineAnalysis:
        """Analyze pipeline and determine optimization strategy.

        Args:
            pipeline: Operation or pipeline to analyze.
            pipeline_context: Optional context about the entire pipeline for global optimization.

        Returns:
            Analysis with execution plan and optimized operation if possible.

        """
        # Flatten pipeline into list of operations
        operations = self._flatten_pipeline(pipeline)

        # Use pipeline_context for global optimization if available
        if pipeline_context:
            # Check if data will return to this storage (enables additional optimizations)
            # This is a placeholder for future global optimizations
            pass

        # Determine what can be executed in SQL
        sql_operations = []
        python_operations = []

        for idx, op in enumerate(operations):
            # Check if operation can be executed in SQL
            # Operations that receive previous_result (idx > 0) may not be optimizable
            has_previous = idx > 0
            can_execute_in_sql = self._can_execute_in_sql(op, has_previous=has_previous)

            if can_execute_in_sql:
                sql_operations.append(op)
            else:
                # If we encounter an operation that cannot be executed in SQL,
                # all subsequent operations are also in Python
                python_operations = list(operations[idx:])
                break

        # Build execution plan
        if not python_operations:
            # Entire pipeline can be executed in SQL
            optimized_op = self._build_optimized_operation(sql_operations)
            return PipelineAnalysis(
                can_execute_on_storage=True,
                optimized_operation=optimized_op,
                execution_plan="storage",
                sql_operations=tuple(sql_operations),
            )

        # Check if we can optimize reado | [filtero]* | updateo/patcho/deleteo pattern
        # This is a special case where operations can be optimized even in pipeline mode
        if sql_operations and python_operations:
            first_python_op = python_operations[0]
            first_python_idx = len(sql_operations)

            # Check if pattern is reado | [filtero]* | updateo/patcho/deleteo
            from haolib.pipelines.operations import (  # noqa: PLC0415
                DeleteOperation,
                PatchOperation,
                ReadOperation,
                UpdateOperation,
            )

            if (
                first_python_idx > 0
                and isinstance(sql_operations[0], ReadOperation)
                and isinstance(first_python_op, (UpdateOperation, PatchOperation, DeleteOperation))
            ):
                # Check if we can optimize this pattern
                can_optimize = self._can_optimize_read_update_pattern(sql_operations, first_python_op)
                if can_optimize:
                    # Can optimize: treat as storage execution plan
                    # The actual optimization will happen in build_optimized_operation_async
                    return PipelineAnalysis(
                        can_execute_on_storage=True,
                        optimized_operation=None,  # Will be built in async method
                        execution_plan="storage",
                        sql_operations=tuple(operations),  # All operations for optimization
                    )

        # Check if first python operation uses previous_result (pipeline mode)
        # If so, entire pipeline must execute in Python (cannot split)
        if python_operations:
            first_python_op = python_operations[0]
            first_python_idx = len(sql_operations)
            uses_previous_result = (
                first_python_idx > 0
                and not self._can_execute_in_sql(first_python_op, has_previous=True)
                and self._operation_uses_previous_result(first_python_op)
            )

            if uses_previous_result:
                # Operation uses previous_result - entire pipeline must be Python
                return PipelineAnalysis(
                    can_execute_on_storage=False,
                    remaining_operations=tuple(operations),
                    execution_plan="python",
                )

        if sql_operations:
            # Hybrid approach: part in SQL, part in Python
            optimized_op = self._build_optimized_operation(sql_operations)
            return PipelineAnalysis(
                can_execute_on_storage=False,
                optimized_operation=optimized_op,
                remaining_operations=tuple(python_operations),
                execution_plan="hybrid",
                sql_operations=tuple(sql_operations),
            )

        # All operations need to be executed in Python
        return PipelineAnalysis(
            can_execute_on_storage=False,
            remaining_operations=tuple(operations),
            execution_plan="python",
        )

    def optimize(
        self, pipeline: Operation[Any, Any] | Pipeline[Any, Any, Any]
    ) -> Operation[Any, Any] | Pipeline[Any, Any, Any]:
        """Optimize pipeline to single SQL query when possible.

        Args:
            pipeline: Operation or pipeline to optimize.

        Returns:
            Optimized operation (may be a single SQL query) or original pipeline.

        """
        analysis = self.analyze(pipeline)
        if analysis.optimized_operation:
            return analysis.optimized_operation
        return pipeline

    def _flatten_pipeline(self, operation: Operation[Any, Any] | Pipeline[Any, Any, Any]) -> list[Operation[Any, Any]]:
        """Flatten pipeline into list of operations.

        Args:
            operation: Operation or pipeline to flatten.

        Returns:
            List of operations in execution order.

        """
        operations: list[Operation] = []
        # Handle TargetBoundOperation and TargetSwitch at the top level
        if isinstance(operation, TargetBoundOperation):
            # Unwrap TargetBoundOperation - flatten its operation
            operation = operation.operation
        elif isinstance(operation, TargetSwitch):
            msg = "TargetSwitch should not be in optimizer"
            raise TypeError(msg)
        current: Operation | Pipeline = operation

        while True:
            if isinstance(current, Pipeline):
                # Add first operation
                if isinstance(current.first, Pipeline):
                    # Recursively flatten nested pipeline
                    operations.extend(self._flatten_pipeline(current.first))
                else:
                    # current.first can be Operation, TargetBoundOperation, or TargetSwitch
                    # Extract operation from TargetBoundOperation or TargetSwitch
                    first_op = current.first
                    if isinstance(first_op, TargetBoundOperation) or isinstance(first_op, TargetSwitch):
                        first_op = first_op.operation
                    operations.append(first_op)

                # Move to second operation
                # current.second can be Operation, Pipeline, TargetBoundOperation, or TargetSwitch
                # Extract operation from TargetBoundOperation or TargetSwitch
                second_op = current.second
                if isinstance(second_op, TargetBoundOperation) or isinstance(second_op, TargetSwitch):
                    second_op = second_op.operation
                # Now second_op is Operation or Pipeline, safe to assign
                current = second_op
            else:
                # This is a simple operation
                operations.append(current)
                break

        return operations

    def _can_execute_in_sql(self, operation: Operation, has_previous: bool = False) -> bool:
        """Check if operation can be executed in SQL.

        Args:
            operation: Operation to check.
            has_previous: Whether operation will receive previous_result from pipeline.

        Returns:
            True if operation can be executed in SQL, False otherwise.

        """
        # Import operations lazily to avoid circular import
        from haolib.pipelines.operations import (  # noqa: PLC0415
            CreateOperation,
            DeleteOperation,
            PatchOperation,
            ReadOperation,
            UpdateOperation,
        )

        # ReadOperation can always be executed in SQL (it never receives previous_result)
        if isinstance(operation, ReadOperation):
            return True

        # CRUD operations can be executed in SQL in two cases:
        # 1. Search mode: have search_index (and data/patch for update/patch)
        # 2. Pipeline mode: receive previous_result from reado in same storage (can be optimized)
        if isinstance(operation, UpdateOperation):
            # UpdateOperation can be in SQL if:
            # - Search mode: has search_index and data
            # - Pipeline mode: receives previous_result (will be optimized with reado)
            if has_previous:
                # Pipeline mode: can be optimized if data is not callable
                # (callable data requires Python execution)
                return not callable(operation.data) if operation.data is not None else False
            # Search mode: must have both search_index and data
            return operation.search_index is not None and operation.data is not None

        if isinstance(operation, PatchOperation):
            # PatchOperation can be in SQL if:
            # - Search mode: has search_index and patch
            # - Pipeline mode: receives previous_result (will be optimized with reado)
            if has_previous:
                # Pipeline mode: can be optimized if patch is not callable
                # (patcho doesn't support callable, but check anyway)
                return not callable(operation.patch) if operation.patch is not None else False
            # Search mode: must have both search_index and patch
            return operation.search_index is not None and operation.patch is not None

        if isinstance(operation, DeleteOperation):
            # DeleteOperation can be in SQL if:
            # - Search mode: has search_index
            # - Pipeline mode: receives previous_result (will be optimized with reado)
            # Both modes can be optimized
            return True

        if isinstance(operation, CreateOperation):
            # CreateOperation can be in SQL if it has data (not using previous_result)
            # If it receives previous_result, it cannot be optimized
            if has_previous:
                # Pipeline mode: cannot be optimized (uses previous_result)
                return False
            # Search mode: can be optimized if it has data
            return bool(operation.data)

        # FilterOperation can be in SQL if predicate can be converted to specifications
        # But only if it's part of a SQL-optimizable pipeline (after ReadOperation)
        from haolib.pipelines.operations import FilterOperation  # noqa: PLC0415

        if isinstance(operation, FilterOperation):
            # FilterOperation always needs previous_result, so it can only be optimized
            # if it's part of a SQL pipeline (after ReadOperation that was optimized)
            return self._predicate_analyzer.can_convert_to_sql(operation)

        # MapOperation, ReduceOperation, TransformOperation need detailed analysis
        # For now return False - detailed analysis will be in next stage
        from haolib.pipelines.operations import (  # noqa: PLC0415
            MapOperation,
            ReduceOperation,
            TransformOperation,
        )

        if isinstance(operation, (MapOperation, ReduceOperation, TransformOperation)):
            return False

        # Unknown operation - execute in Python
        return False

    def _operation_uses_previous_result(self, operation: Operation) -> bool:
        """Check if operation uses previous_result (pipeline mode).

        Args:
            operation: Operation to check.

        Returns:
            True if operation uses previous_result instead of search_index/data.

        """
        from haolib.pipelines.operations import (  # noqa: PLC0415
            DeleteOperation,
            PatchOperation,
            UpdateOperation,
        )

        if isinstance(operation, UpdateOperation):
            # Uses previous_result if no search_index or data
            return operation.search_index is None or operation.data is None

        if isinstance(operation, PatchOperation):
            # Uses previous_result if no search_index or patch
            return operation.search_index is None or operation.patch is None

        if isinstance(operation, DeleteOperation):
            # Uses previous_result if no search_index
            return operation.search_index is None

        # Other operations don't use previous_result in this way
        return False

    def _can_optimize_read_update_pattern(self, sql_operations: list[Operation], update_op: Operation) -> bool:
        """Check if reado | [filtero]* | updateo/patcho/deleteo pattern can be optimized.

        Args:
            sql_operations: List of SQL operations (should start with ReadOperation).
            update_op: Update/Patch/Delete operation to check.

        Returns:
            True if pattern can be optimized, False otherwise.

        """
        from haolib.pipelines.operations import (  # noqa: PLC0415
            DeleteOperation,
            FilterOperation,
            PatchOperation,
            ReadOperation,
            UpdateOperation,
        )

        # First operation must be ReadOperation
        if not sql_operations or not isinstance(sql_operations[0], ReadOperation):
            return False

        # Check if all operations between read and update are FilterOperations
        for op in sql_operations[1:]:
            if not isinstance(op, FilterOperation):
                return False

        # Check type compatibility
        # For UpdateOperation, PatchOperation, DeleteOperation, we can't directly
        # check data_type since they don't have it. Instead, we rely on the fact
        # that if they're in the same pipeline, they should work with the same type.
        # The actual type check happens in QueryBuilder when building the query.
        if isinstance(update_op, UpdateOperation):
            # Check if data is callable (cannot optimize callable)
            return not callable(update_op.data) if update_op.data is not None else False

        if isinstance(update_op, PatchOperation):
            # Check if patch is callable (patcho doesn't support callable)
            return not callable(update_op.patch) if update_op.patch is not None else False

        if isinstance(update_op, DeleteOperation):
            # Type compatibility will be checked in QueryBuilder
            return True

        return False

    def _build_optimized_operation(self, operations: list[Operation]) -> Operation | None:
        """Build optimized operation from list of operations.

        Note: This method returns a placeholder. The actual query building
        happens in build_optimized_operation_async() which requires a session.

        Args:
            operations: List of operations that can be executed in SQL.

        Returns:
            Optimized operation (ReadOperation with optimized query) or None.

        """
        if not operations:
            return None

        # Find ReadOperation as base
        from haolib.pipelines.operations import ReadOperation  # noqa: PLC0415

        read_op: ReadOperation[Any] | None = None
        for op in operations:
            if isinstance(op, ReadOperation):
                read_op = op
                break

        if not read_op:
            # If no ReadOperation, cannot optimize
            return None

        # If there are only ReadOperation (no filters), return as is
        if len(operations) == 1:
            return read_op

        # For operations with filters, we need to build query in async method
        # Return a marker that indicates we need to build optimized query
        # We'll store operations for later use
        return read_op

    async def build_optimized_operation_async(self, operations: list[Operation], session: Any) -> Operation | None:
        """Build optimized operation from list of operations (async version).

        This method actually builds the optimized SQL query using QueryBuilder.
        Supports patterns:
        - reado | [filtero]* -> optimized ReadOperation
        - reado | [filtero]* | updateo -> optimized UpdateOperation
        - reado | [filtero]* | patcho -> optimized PatchOperation
        - reado | [filtero]* | deleteo -> optimized DeleteOperation

        Args:
            operations: List of operations that can be executed in SQL.
            session: SQLAlchemy session for building queries.

        Returns:
            Optimized operation or None.

        """
        from haolib.pipelines.operations import (  # noqa: PLC0415
            DeleteOperation,
            FilterOperation,
            PatchOperation,
            ReadOperation,
            UpdateOperation,
        )

        if not operations:
            return None

        # First operation must be ReadOperation
        read_op = operations[0]
        if not isinstance(read_op, ReadOperation):
            return None

        # Separate filter operations from update/patch/delete operations
        filter_ops: list[FilterOperation] = []
        update_op: UpdateOperation[Any] | None = None
        patch_op: PatchOperation[Any] | None = None
        delete_op: DeleteOperation[Any] | None = None

        for op in operations[1:]:
            if isinstance(op, FilterOperation):
                filter_ops.append(op)
            elif isinstance(op, UpdateOperation):
                update_op = op
            elif isinstance(op, PatchOperation):
                patch_op = op
            elif isinstance(op, DeleteOperation):
                delete_op = op

        # Build optimized query based on pattern
        if update_op is not None:
            return await self._query_builder.build_update_async(read_op, filter_ops, update_op, session)

        if patch_op is not None:
            return await self._query_builder.build_patch_async(read_op, filter_ops, patch_op, session)

        if delete_op is not None:
            return await self._query_builder.build_delete_async(read_op, filter_ops, delete_op, session)

        # Default: build optimized ReadOperation (reado | [filtero]*)
        return await self._query_builder.build_async(operations, session)



"""Integration tests for complex pipelines with edge cases and real-world scenarios.

Tests cover:
- Complex multi-storage pipelines (SQL -> S3 -> SQL)
- Edge cases (empty results, single items, large datasets)
- Real-world ETL scenarios
- Data transformation and aggregation
- Multiple storage switches
- Nested pipelines
- Error handling scenarios
"""

from contextlib import suppress
from typing import Any

import pytest
from sqlalchemy import select

from haolib.database.files.s3.clients.abstract import (
    AbstractS3Client,
    S3BucketAlreadyExistsClientException,
    S3BucketAlreadyOwnedByYouClientException,
)
from haolib.database.files.s3.clients.pydantic import S3DeleteObjectsDelete, S3DeleteObjectsDeleteObject
from haolib.pipelines import filtero, mapo, reduceo, transformo
from haolib.storages.data_types.registry import DataTypeRegistry
from haolib.storages.indexes.params import ParamIndex
from haolib.storages.indexes.path import PathIndex
from haolib.storages.indexes.sql import SQLQueryIndex
from haolib.storages.s3 import S3Storage
from haolib.storages.sqlalchemy import SQLAlchemyStorage
from tests.integration.storages.conftest import User, UserModel

# Additional S3 buckets for complex multi-storage pipelines
RAW_DATA_BUCKET = "test-raw-data-bucket"
PROCESSED_DATA_BUCKET = "test-processed-data-bucket"
ARCHIVE_BUCKET = "test-archive-bucket"


class TestComplexPipelines:
    """Test complex pipelines with multiple operations and storage switches."""

    @pytest.mark.asyncio
    async def test_etl_pipeline_sql_to_s3_to_sql(
        self,
        sqlalchemy_storage: SQLAlchemyStorage,
        s3_storage: S3Storage,
    ) -> None:
        """Test ETL pipeline: SQL -> S3 -> SQL."""
        # Create initial data in SQL
        users = [
            User(name="Alice", age=25, email="alice@example.com"),
            User(name="Bob", age=30, email="bob@example.com"),
            User(name="Charlie", age=35, email="charlie@example.com"),
        ]

        # Create in SQL -> Transform -> Save to S3 -> Count (avoiding duplicate creation)
        result = await (
            sqlalchemy_storage.create(users).returning()
            | mapo(lambda user, _idx: {"name": user.name, "age": user.age, "email": user.email})
            | s3_storage.create()
            | mapo(lambda data, _idx: data[0])  # Extract dict from tuple
            | reduceo(lambda acc, d: acc + 1, 0)
        ).execute()

        assert result == 3

    @pytest.mark.asyncio
    async def test_aggregation_pipeline(
        self,
        sqlalchemy_storage: SQLAlchemyStorage,
    ) -> None:
        """Test aggregation pipeline with filter, map, and reduce."""
        users = [
            User(name="Alice", age=25, email="alice@example.com"),
            User(name="Bob", age=30, email="bob@example.com"),
            User(name="Charlie", age=35, email="charlie@example.com"),
            User(name="David", age=40, email="david@example.com"),
        ]

        # Create -> Filter (age >= 30) -> Map (extract age) -> Reduce (sum)
        result = await (
            sqlalchemy_storage.create(users).returning()
            | filtero(lambda user: user.age >= 30)
            | mapo(lambda user, _idx: user.age)
            | reduceo(lambda acc, age: acc + age, 0)
        ).execute()

        assert result == 105  # 30 + 35 + 40

    @pytest.mark.asyncio
    async def test_multiple_storage_switches(
        self,
        sqlalchemy_storage: SQLAlchemyStorage,
        s3_storage: S3Storage,
    ) -> None:
        """Test pipeline with multiple storage switches."""
        users = [
            User(name="Alice", age=25, email="alice@example.com"),
            User(name="Bob", age=30, email="bob@example.com"),
        ]

        # SQL -> S3 (test multiple storage switches by creating separately)
        # First create in SQL
        await sqlalchemy_storage.create(users).returning().execute()

        # Then create in S3 (returns tuples)
        result = await s3_storage.create(users).returning().execute()

        assert len(result) == 2
        assert all(isinstance(item, tuple) and len(item) == 2 for item in result)

    @pytest.mark.asyncio
    async def test_empty_result_pipeline(
        self,
        sqlalchemy_storage: SQLAlchemyStorage,
    ) -> None:
        """Test pipeline with empty results."""
        # Read non-existent data -> Filter -> Map -> Reduce
        result = await (
            sqlalchemy_storage.read(ParamIndex(User, email="nonexistent@example.com")).returning()
            | filtero(lambda user: user.age > 20)
            | mapo(lambda user, _idx: user.name)
            | reduceo(lambda acc, name: acc + name, "")
        ).execute()

        assert result == ""

    @pytest.mark.asyncio
    async def test_single_item_pipeline(
        self,
        sqlalchemy_storage: SQLAlchemyStorage,
    ) -> None:
        """Test pipeline with single item."""
        user = User(name="Alice", age=25, email="alice@example.com")

        result = await (
            sqlalchemy_storage.create([user]).returning()
            | filtero(lambda u: u.age > 20)
            | mapo(lambda u, _idx: u.name.upper())
            | reduceo(lambda acc, name: acc + name, "")
        ).execute()

        assert result == "ALICE"

    @pytest.mark.asyncio
    async def test_large_dataset_pipeline(
        self,
        sqlalchemy_storage: SQLAlchemyStorage,
    ) -> None:
        """Test pipeline with large dataset."""
        # Create 100 users
        users = [User(name=f"User{i}", age=20 + (i % 30), email=f"user{i}@example.com") for i in range(100)]

        # Create -> Filter (age >= 25) -> Map (age) -> Reduce (sum)
        result = await (
            sqlalchemy_storage.create(users).returning()
            | filtero(lambda user: user.age >= 25)
            | mapo(lambda user, _idx: user.age)
            | reduceo(lambda acc, age: acc + age, 0)
        ).execute()

        # Verify result is reasonable (sum of ages >= 25)
        assert result > 0
        assert isinstance(result, int)

    @pytest.mark.asyncio
    async def test_nested_transformations(
        self,
        sqlalchemy_storage: SQLAlchemyStorage,
    ) -> None:
        """Test nested transformation operations."""
        users = [
            User(name="Alice", age=25, email="alice@example.com"),
            User(name="Bob", age=30, email="bob@example.com"),
            User(name="Charlie", age=35, email="charlie@example.com"),
        ]

        # Multiple filters and maps
        result = await (
            sqlalchemy_storage.create(users).returning()
            | filtero(lambda u: u.age >= 25)
            | mapo(lambda u, _idx: {"name": u.name, "age": u.age})
            | filtero(lambda d: len(d["name"]) >= 3)  # Changed to >= to include BOB
            | mapo(lambda d, _idx: d["name"].upper())
            | reduceo(lambda acc, name: f"{acc},{name}" if acc else name, "")
        ).execute()

        assert "ALICE" in result
        assert "BOB" in result
        assert "CHARLIE" in result

    @pytest.mark.asyncio
    async def test_update_after_read_pipeline(
        self,
        sqlalchemy_storage: SQLAlchemyStorage,
    ) -> None:
        """Test read -> transform -> update pipeline."""
        # Create initial data
        users = [
            User(name="Alice", age=25, email="alice@example.com"),
            User(name="Bob", age=30, email="bob@example.com"),
        ]

        await sqlalchemy_storage.create(users).returning().execute()

        # Read -> Transform -> Update (use patcho for partial update)
        result = await (
            sqlalchemy_storage.read(ParamIndex(User, email="alice@example.com")).returning()
            | sqlalchemy_storage.patch({"age": 26})
        ).execute()

        assert len(result) == 1
        assert result[0].age == 26

    @pytest.mark.asyncio
    async def test_create_without_data_uses_previous_result(
        self,
        sqlalchemy_storage: SQLAlchemyStorage,
    ) -> None:
        """Test create() without data uses previous_result."""
        users = [
            User(name="Alice", age=25, email="alice@example.com"),
            User(name="Bob", age=30, email="bob@example.com"),
        ]

        # Create -> Filter -> Count (without creating again to avoid unique constraint)
        result = await (
            sqlalchemy_storage.create(users).returning()
            | filtero(lambda u: u.age >= 30)
            | reduceo(lambda acc, u: acc + 1, 0)
        ).execute()

        assert result == 1  # Only Bob

    @pytest.mark.asyncio
    async def test_create_pass_through_pipeline(
        self,
        sqlalchemy_storage: SQLAlchemyStorage,
    ) -> None:
        """Test create() as pass-through operation."""
        users = [
            User(name="Alice", age=25, email="alice@example.com"),
            User(name="Bob", age=30, email="bob@example.com"),
        ]

        # Create -> Pass through create() -> Filter -> Count
        result = await (
            sqlalchemy_storage.create(users).returning()
            | sqlalchemy_storage.create()  # Pass through without saving
            | filtero(lambda u: u.age >= 30)
            | reduceo(lambda acc, u: acc + 1, 0)
        ).execute()

        assert result == 1  # Only Bob

    @pytest.mark.asyncio
    async def test_complex_etl_with_multiple_transforms(
        self,
        sqlalchemy_storage: SQLAlchemyStorage,
        s3_storage: S3Storage,
    ) -> None:
        """Test complex ETL with multiple transformations."""
        users = [
            User(name="Alice", age=25, email="alice@example.com"),
            User(name="Bob", age=30, email="bob@example.com"),
            User(name="Charlie", age=35, email="charlie@example.com"),
        ]

        # SQL -> Transform -> S3 -> Transform -> Filter -> Count
        result = await (
            sqlalchemy_storage.create(users).returning()
            | mapo(lambda u, _idx: {"name": u.name.upper(), "age": u.age, "email": u.email})
            | s3_storage.create()
            | mapo(lambda data, _idx: data[0])  # Extract dict from tuple
            | mapo(lambda d, _idx: d["age"])  # Extract age
            | filtero(lambda age: age >= 30)
            | reduceo(lambda acc, age: acc + 1, 0)
        ).execute()

        assert result == 2  # Bob and Charlie

    @pytest.mark.asyncio
    async def test_read_update_delete_pipeline(
        self,
        sqlalchemy_storage: SQLAlchemyStorage,
    ) -> None:
        """Test read -> update -> delete pipeline."""
        users = [
            User(name="Alice", age=25, email="alice@example.com"),
            User(name="Bob", age=30, email="bob@example.com"),
        ]

        # Create -> Read -> Update -> Delete
        await sqlalchemy_storage.create(users).returning().execute()

        deleted_count = await (
            sqlalchemy_storage.read(ParamIndex(User, email="alice@example.com")).returning()
            | sqlalchemy_storage.patch({"age": 26})
            | sqlalchemy_storage.delete()
        ).execute()

        assert deleted_count == 1

    @pytest.mark.asyncio
    async def test_filter_chain_pipeline(
        self,
        sqlalchemy_storage: SQLAlchemyStorage,
    ) -> None:
        """Test multiple filters in chain."""
        users = [
            User(name="Alice", age=25, email="alice@example.com"),
            User(name="Bob", age=30, email="bob@example.com"),
            User(name="Charlie", age=35, email="charlie@example.com"),
            User(name="David", age=40, email="david@example.com"),
        ]

        # Create -> Filter (age >= 30) -> Filter (name starts with 'C' or 'D') -> Count
        result = await (
            sqlalchemy_storage.create(users).returning()
            | filtero(lambda u: u.age >= 30)
            | filtero(lambda u: u.name.startswith("C") or u.name.startswith("D"))
            | reduceo(lambda acc, u: acc + 1, 0)
        ).execute()

        assert result == 2  # Charlie and David

    @pytest.mark.asyncio
    async def test_map_chain_pipeline(
        self,
        sqlalchemy_storage: SQLAlchemyStorage,
    ) -> None:
        """Test multiple maps in chain."""
        users = [
            User(name="Alice", age=25, email="alice@example.com"),
            User(name="Bob", age=30, email="bob@example.com"),
        ]

        # Create -> Map (name) -> Map (upper) -> Map (reverse) -> Reduce
        result = await (
            sqlalchemy_storage.create(users).returning()
            | mapo(lambda u, _idx: u.name)
            | mapo(lambda name, _idx: name.upper())
            | mapo(lambda name, _idx: name[::-1])
            | reduceo(lambda acc, name: f"{acc},{name}" if acc else name, "")
        ).execute()

        assert "ECILA" in result
        assert "BOB" in result

    @pytest.mark.asyncio
    async def test_transform_with_side_effects(
        self,
        sqlalchemy_storage: SQLAlchemyStorage,
    ) -> None:
        """Test transform operation with complex logic."""
        users = [
            User(name="Alice", age=25, email="alice@example.com"),
            User(name="Bob", age=30, email="bob@example.com"),
        ]

        # Create -> Transform (calculate statistics)
        result = await (
            sqlalchemy_storage.create(users).returning()
            | transformo(
                lambda users: {
                    "count": len(users),
                    "total_age": sum(u.age for u in users),
                    "avg_age": sum(u.age for u in users) / len(users) if users else 0,
                }
            )
        ).execute()

        assert result["count"] == 2
        assert result["total_age"] == 55
        assert result["avg_age"] == 27.5

    @pytest.mark.asyncio
    async def test_sql_query_index_pipeline(
        self,
        sqlalchemy_storage: SQLAlchemyStorage,
    ) -> None:
        """Test pipeline with SQLQueryIndex."""
        users = [
            User(name="Alice", age=25, email="alice@example.com"),
            User(name="Bob", age=30, email="bob@example.com"),
            User(name="Charlie", age=35, email="charlie@example.com"),
        ]

        await sqlalchemy_storage.create(users).returning().execute()

        # Read with SQLQueryIndex -> Filter -> Map
        result = await (
            sqlalchemy_storage.read(SQLQueryIndex(query=select(UserModel).where(UserModel.age >= 30))).returning()
            | filtero(lambda u: len(u.name) >= 3)  # Changed to >= to include Bob
            | mapo(lambda u, _idx: u.name)
        ).execute()

        assert "Bob" in result
        assert "Charlie" in result
        assert len(result) == 2

    @pytest.mark.asyncio
    async def test_update_with_sql_query_index(
        self,
        sqlalchemy_storage: SQLAlchemyStorage,
    ) -> None:
        """Test update with SQLQueryIndex."""
        users = [
            User(name="Alice", age=25, email="alice@example.com"),
            User(name="Bob", age=30, email="bob@example.com"),
        ]

        await sqlalchemy_storage.create(users).returning().execute()

        # Update with SQLQueryIndex (use patcho for partial update)
        result = (
            await sqlalchemy_storage.read(SQLQueryIndex(query=select(UserModel).where(UserModel.age >= 30)))
            .patch({"age": 31})
            .returning()
            .execute()
        )

        assert len(result) == 1
        assert result[0].age == 31
        assert result[0].name == "Bob"

    @pytest.mark.asyncio
    async def test_delete_with_sql_query_index(
        self,
        sqlalchemy_storage: SQLAlchemyStorage,
    ) -> None:
        """Test delete with SQLQueryIndex."""
        users = [
            User(name="Alice", age=25, email="alice@example.com"),
            User(name="Bob", age=30, email="bob@example.com"),
            User(name="Charlie", age=35, email="charlie@example.com"),
        ]

        await sqlalchemy_storage.create(users).returning().execute()

        # Delete with SQLQueryIndex
        deleted_count = (
            await sqlalchemy_storage.read(SQLQueryIndex(query=select(UserModel).where(UserModel.age >= 30)))
            .delete()
            .execute()
        )

        assert deleted_count == 2  # Bob and Charlie

    @pytest.mark.asyncio
    async def test_s3_create_returns_tuples_pipeline(
        self,
        sqlalchemy_storage: SQLAlchemyStorage,
        s3_storage: S3Storage,
    ) -> None:
        """Test that S3 create returns tuples and they are handled correctly."""
        users = [
            User(name="Alice", age=25, email="alice@example.com"),
            User(name="Bob", age=30, email="bob@example.com"),
        ]

        # SQL -> S3 (returns tuples) -> Extract data -> Count
        # First create in SQL
        await sqlalchemy_storage.create(users).returning().execute()

        # Then create in S3 (returns tuples)
        s3_result = await s3_storage.create(users).returning().execute()

        # Extract data from tuples and count
        result = len([data[0] for data in s3_result])
        assert result == 2

    @pytest.mark.asyncio
    async def test_empty_create_pipeline(
        self,
        sqlalchemy_storage: SQLAlchemyStorage,
    ) -> None:
        """Test create with empty list."""
        result = await sqlalchemy_storage.create([]).returning().execute()

        assert result == []

    @pytest.mark.asyncio
    async def test_reduce_with_empty_list(
        self,
        sqlalchemy_storage: SQLAlchemyStorage,
    ) -> None:
        """Test reduce with empty list."""
        # Read non-existent -> Reduce
        result = await (
            sqlalchemy_storage.read(ParamIndex(User, email="nonexistent@example.com")).returning()
            | reduceo(lambda acc, u: acc + u.age, 0)
        ).execute()

        assert result == 0

    @pytest.mark.asyncio
    async def test_complex_nested_pipeline(
        self,
        sqlalchemy_storage: SQLAlchemyStorage,
        s3_storage: S3Storage,
    ) -> None:
        """Test deeply nested pipeline with multiple operations."""
        users = [
            User(name="Alice", age=25, email="alice@example.com"),
            User(name="Bob", age=30, email="bob@example.com"),
            User(name="Charlie", age=35, email="charlie@example.com"),
            User(name="David", age=40, email="david@example.com"),
        ]

        # Very complex pipeline: Create -> Filter -> Map -> S3 -> Map -> Filter -> Map -> Reduce
        result = await (
            sqlalchemy_storage.create(users).returning()
            | filtero(lambda u: u.age >= 30)
            | mapo(lambda u, _idx: {"name": u.name, "age": u.age})
            | s3_storage.create()
            | mapo(lambda data, _idx: data[0])  # Extract dict from tuple
            | mapo(lambda d, _idx: d["age"])  # Extract age
            | filtero(lambda age: age >= 35)
            | reduceo(lambda acc, age: acc + age, 0)
        ).execute()

        assert result == 75  # 35 + 40

    @pytest.mark.asyncio
    async def test_pipeline_with_callable_data(
        self,
        sqlalchemy_storage: SQLAlchemyStorage,
    ) -> None:
        """Test createo with callable data."""
        users = [
            User(name="Alice", age=25, email="alice@example.com"),
            User(name="Bob", age=30, email="bob@example.com"),
        ]

        # Create initial data
        await sqlalchemy_storage.create(users).returning().execute()

        # Read -> Update with callable (increment age)
        # Use patcho for partial update
        result = await (
            sqlalchemy_storage.read(ParamIndex(User, email="alice@example.com")).returning()
            | sqlalchemy_storage.patch({"age": 26})
        ).execute()

        assert len(result) == 1
        assert result[0].age == 26

    @pytest.mark.asyncio
    async def test_multiple_reads_pipeline(
        self,
        sqlalchemy_storage: SQLAlchemyStorage,
    ) -> None:
        """Test pipeline with multiple read operations."""
        users = [
            User(name="Alice", age=25, email="alice@example.com"),
            User(name="Bob", age=30, email="bob@example.com"),
            User(name="Charlie", age=35, email="charlie@example.com"),
        ]

        await sqlalchemy_storage.create(users).returning().execute()

        # Read Alice -> Read Bob -> Combine
        result = await (
            sqlalchemy_storage.read(ParamIndex(User, email="alice@example.com")).returning()
            | sqlalchemy_storage.create()  # Pass through
            | sqlalchemy_storage.read(ParamIndex(User, email="bob@example.com")).returning()
            | reduceo(lambda acc, u: acc + [u], [])
        ).execute()

        assert len(result) == 1  # Only Bob (Alice was passed through but not combined)
        assert result[0].name == "Bob"

    @pytest.mark.asyncio
    async def test_pipeline_with_all_operations(
        self,
        sqlalchemy_storage: SQLAlchemyStorage,
    ) -> None:
        """Test pipeline using all operation types."""
        users = [
            User(name="Alice", age=25, email="alice@example.com"),
            User(name="Bob", age=30, email="bob@example.com"),
        ]

        # Create -> Read -> Filter -> Map -> Transform -> Reduce -> Update -> Delete
        await sqlalchemy_storage.create(users).returning().execute()

        # Read -> Filter -> Map -> Transform -> Update
        result = await (
            sqlalchemy_storage.read(ParamIndex(User, email="alice@example.com")).returning()
            | filtero(lambda u: u.age > 20)
            | transformo(lambda users: users[0] if users else None)
        ).execute()

        assert result is not None
        assert result.age == 25
        # Update separately (use patcho for partial update)
        updated = (
            await sqlalchemy_storage.read(ParamIndex(User, email="alice@example.com"))
            .patch({"age": 26})
            .returning()
            .execute()
        )
        assert updated[0].age == 26

    @pytest.mark.asyncio
    async def test_pipeline_with_conditional_logic(
        self,
        sqlalchemy_storage: SQLAlchemyStorage,
    ) -> None:
        """Test pipeline with conditional transformations."""
        users = [
            User(name="Alice", age=25, email="alice@example.com"),
            User(name="Bob", age=30, email="bob@example.com"),
            User(name="Charlie", age=35, email="charlie@example.com"),
        ]

        # Create -> Filter -> Map (conditional) -> Filter -> Count
        result = await (
            sqlalchemy_storage.create(users).returning()
            | filtero(lambda u: u.age >= 30)
            | mapo(lambda u, _idx: u.name.upper() if u.age >= 35 else u.name)
            | filtero(lambda name: name.isupper())
            | reduceo(lambda acc, name: acc + 1, 0)
        ).execute()

        assert result == 1  # Only CHARLIE

    @pytest.mark.asyncio
    async def test_pipeline_with_string_operations(
        self,
        sqlalchemy_storage: SQLAlchemyStorage,
    ) -> None:
        """Test pipeline with string transformations."""
        users = [
            User(name="Alice", age=25, email="alice@example.com"),
            User(name="Bob", age=30, email="bob@example.com"),
        ]

        # Create -> Map (name) -> Map (upper) -> Map (reverse) -> Reduce (join)
        result = await (
            sqlalchemy_storage.create(users).returning()
            | mapo(lambda u, _idx: u.name)
            | mapo(lambda name, _idx: name.upper())
            | mapo(lambda name, _idx: name[::-1])
            | reduceo(lambda acc, name: f"{acc}|{name}" if acc else name, "")
        ).execute()

        assert "ECILA" in result
        assert "BOB" in result

    @pytest.mark.asyncio
    async def test_pipeline_with_numeric_operations(
        self,
        sqlalchemy_storage: SQLAlchemyStorage,
    ) -> None:
        """Test pipeline with numeric transformations."""
        users = [
            User(name="Alice", age=25, email="alice@example.com"),
            User(name="Bob", age=30, email="bob@example.com"),
            User(name="Charlie", age=35, email="charlie@example.com"),
        ]

        # Create -> Map (age) -> Reduce (sum) -> Transform (calculate stats)
        total = await (
            sqlalchemy_storage.create(users).returning()
            | mapo(lambda u, _idx: u.age)
            | reduceo(lambda acc, age: acc + age, 0)
        ).execute()

        assert total == 90

        # Calculate average
        avg = await (
            sqlalchemy_storage.read(ParamIndex(User)).returning()
            | mapo(lambda u, _idx: u.age)
            | transformo(lambda ages: sum(ages) / len(ages) if ages else 0)
        ).execute()

        assert avg == 30.0

    @pytest.mark.asyncio
    async def test_pipeline_with_error_recovery(
        self,
        sqlalchemy_storage: SQLAlchemyStorage,
    ) -> None:
        """Test pipeline that handles edge cases gracefully."""
        users = [
            User(name="Alice", age=25, email="alice@example.com"),
            User(name="Bob", age=30, email="bob@example.com"),
        ]

        # Create -> Filter (always true) -> Map -> Filter (always false) -> Reduce
        result = await (
            sqlalchemy_storage.create(users).returning()
            | filtero(lambda u: True)  # All pass
            | mapo(lambda u, _idx: u.name)
            | filtero(lambda name: False)  # None pass
            | reduceo(lambda acc, name: acc + name, "")
        ).execute()

        assert result == ""

    @pytest.mark.asyncio
    async def test_pipeline_with_duplicate_operations(
        self,
        sqlalchemy_storage: SQLAlchemyStorage,
    ) -> None:
        """Test pipeline with duplicate operations."""
        users = [
            User(name="Alice", age=25, email="alice@example.com"),
            User(name="Bob", age=30, email="bob@example.com"),
        ]

        # Create -> Filter -> Filter -> Filter (same condition)
        result = await (
            sqlalchemy_storage.create(users).returning()
            | filtero(lambda u: u.age >= 30)
            | filtero(lambda u: u.age >= 30)
            | filtero(lambda u: u.age >= 30)
            | reduceo(lambda acc, u: acc + 1, 0)
        ).execute()

        assert result == 1  # Only Bob

    @pytest.mark.asyncio
    async def test_pipeline_with_opposite_filters(
        self,
        sqlalchemy_storage: SQLAlchemyStorage,
    ) -> None:
        """Test pipeline with opposite filter conditions."""
        users = [
            User(name="Alice", age=25, email="alice@example.com"),
            User(name="Bob", age=30, email="bob@example.com"),
            User(name="Charlie", age=35, email="charlie@example.com"),
        ]

        # Create -> Filter (age >= 30) -> Filter (age < 35) -> Count
        result = await (
            sqlalchemy_storage.create(users).returning()
            | filtero(lambda u: u.age >= 30)
            | filtero(lambda u: u.age < 35)
            | reduceo(lambda acc, u: acc + 1, 0)
        ).execute()

        assert result == 1  # Only Bob (30 <= age < 35)

    @pytest.mark.asyncio
    async def test_pipeline_with_mixed_data_types(
        self,
        sqlalchemy_storage: SQLAlchemyStorage,
        s3_storage: S3Storage,
    ) -> None:
        """Test pipeline handling mixed data types."""
        users = [
            User(name="Alice", age=25, email="alice@example.com"),
            User(name="Bob", age=30, email="bob@example.com"),
        ]

        # SQL (User objects) -> Map (dict) -> S3 (dict) -> Map (extract dict) -> Count
        result = await (
            sqlalchemy_storage.create(users).returning()
            | mapo(lambda u, _idx: {"name": u.name, "age": u.age, "email": u.email})
            | s3_storage.create()
            | mapo(lambda data, _idx: data[0])  # Extract dict from tuple
            | reduceo(lambda acc, d: acc + 1, 0)
        ).execute()

        assert result == 2



"""Abstract database transactions."""

from types import TracebackType
from typing import Any, Protocol, Self


class AbstractDatabaseTransaction[DBManipulator: Any](Protocol):
    """Abstract database transaction."""

    async def __aenter__(self) -> Self:
        """Enter the transaction."""
        ...

    async def __aexit__(
        self, exc_type: type[BaseException] | None, exc_value: BaseException | None, traceback: TracebackType | None
    ) -> None:
        """Leave the transaction."""
        ...

    async def get_manipulator(self) -> DBManipulator:
        """Get the database manipulator."""
        ...


class DatabaseTransactionError(Exception):
    """Database transaction error."""



"""Observability config."""

import os

from pydantic import Field
from pydantic_settings import BaseSettings


def get_enable_otel_tracer() -> bool:
    """Get if otel tracer is enabled."""
    return "OTEL_EXPORTER_OTLP_ENDPOINT" in os.environ


class ObservabilityConfig(BaseSettings):
    """Observability configuration.

    This config is used to configure the observability.

    Attributes:
        service_namespace (str): The namespace of the service.
            Defaults to "Humanlessly Autonomously Orchestrated Service".
        enable_otel_tracer (bool): Whether to enable the otel tracer.
            Defaults to the value of the "OTEL_EXPORTER_OTLP_ENDPOINT" environment variable.
        enable_console_tracer (bool): Whether to enable the console tracer. Defaults to False.
        enable_otel_metrics (bool): Whether to enable the otel metrics.
            Defaults to the value of the "OTEL_EXPORTER_OTLP_ENDPOINT" environment variable.
        enable_console_metrics (bool): Whether to enable the console metrics. Defaults to False.
        enable_otel_logs (bool): Whether to enable the otel logs.
            Defaults to the value of the "OTEL_EXPORTER_OTLP_ENDPOINT" environment variable.
        enable_console_logs (bool): Whether to enable the console logs. Defaults to False.
        suppress_httpx_logs (bool): Whether to suppress the httpx logs. Defaults to True.

    """

    service_namespace: str = "Humanlessly Autonomously Orchestrated Service"

    enable_otel_tracer: bool = Field(
        default_factory=get_enable_otel_tracer, description="Whether to enable the otel tracer."
    )
    enable_console_tracer: bool = Field(default=False, description="Whether to enable the console tracer.")

    enable_otel_metrics: bool = Field(
        default_factory=get_enable_otel_tracer, description="Whether to enable the otel metrics."
    )
    enable_console_metrics: bool = Field(default=False, description="Whether to enable the console metrics.")

    enable_otel_logs: bool = Field(
        default_factory=get_enable_otel_tracer, description="Whether to enable the otel logs."
    )
    enable_console_logs: bool = Field(default=False, description="Whether to enable the console logs.")

    suppress_httpx_logs: bool = Field(default=True, description="Whether to suppress the httpx logs.")



"""Default exception handlers for FastAPI application."""

import logging

from fastapi import Request
from fastapi.responses import JSONResponse

from haolib.exceptions.base.fastapi import FastAPIBaseException
from haolib.exceptions.schemas.fastapi import FastAPIErrorSchema
from haolib.observability.utils import observe_exception

logger = logging.getLogger(__name__)


async def fastapi_base_exception_handler(
    request: Request,
    exc: FastAPIBaseException,
) -> JSONResponse:
    """Exception handler for FastAPIBaseException.

    Args:
        request: The request.
        exc: The exception.

    Returns:
        JSONResponse: JSON serialized ErrorModel.

    """

    error_schema = FastAPIErrorSchema(
        error_code=exc.current_error_code,
        detail=exc.current_detail,
        additional_info=exc.current_additional_info,
    ).model_dump(mode="json")

    return JSONResponse(
        status_code=exc.current_status_code,
        content=error_schema,
        headers=exc.current_headers,
    )


async def fastapi_unknown_exception_handler(request: Request, exc: Exception) -> JSONResponse:
    """Exception handler for unknown exceptions.

    Args:
        request: The request.
        exc: The exception.

    Returns:
        JSONResponse: JSON serialized ErrorModel.

    """
    logger.exception("Unknown exception occurred. Details:", exc_info=exc)

    error_schema = FastAPIErrorSchema(
        error_code="UNKNOWN_EXCEPTION",
        detail="Unknown exception occurred",
        additional_info={},
    ).model_dump(mode="json")

    return JSONResponse(status_code=500, content=error_schema)


async def fastapi_unknown_exception_handler_with_observability(request: Request, exc: Exception) -> JSONResponse:
    """Exception handler for unknown exceptions with observability.

    Args:
        request: The request.
        exc: The exception.

    Returns:
        JSONResponse: JSON serialized ErrorModel.

    """
    await observe_exception(exc)
    return await fastapi_unknown_exception_handler(request, exc)



"""Pipeline optimizers for storage backends.

This module re-exports optimizer types from haolib.pipelines
for backward compatibility.
New code should import from haolib.pipelines directly.
"""

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from haolib.pipelines import PipelineAnalysis, PipelineOptimizer


# For runtime, import lazily to avoid circular import
def __getattr__(name: str) -> Any:
    """Lazy import for backward compatibility."""
    from haolib import pipelines  # noqa: PLC0415

    return getattr(pipelines, name)


__all__ = [
    "PipelineAnalysis",
    "PipelineOptimizer",
]



"""FastMCP entrypoint."""

import logging
from typing import TYPE_CHECKING, Any, Self

from haolib.components.events import EventEmitter
from haolib.components.plugins.helpers import apply_plugin, apply_preset
from haolib.components.plugins.registry import PluginRegistry
from haolib.entrypoints.abstract import (
    AbstractEntrypoint,
)
from haolib.entrypoints.events.abstract import EntrypointShutdownEvent, EntrypointStartupEvent
from haolib.entrypoints.plugins.abstract import AbstractEntrypointPlugin, AbstractEntrypointPluginPreset

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from fastmcp import FastMCP


class FastMCPEntrypoint(AbstractEntrypoint):
    """FastMCP entrypoint implementation with plugin system.

    Provides a builder-pattern interface for configuring and running FastMCP
    applications using plugins for common features like exception handling.

    Example:
        ```python
        from fastmcp import FastMCP
        from haolib.entrypoints.plugins.fastmcp import FastMCPExceptionHandlersPlugin

        fastmcp = FastMCP()
        entrypoint = (
            FastMCPEntrypoint(fastmcp=fastmcp)
            .use_plugin(FastMCPExceptionHandlersPlugin({
                ValueError: lambda exc, ctx: logger.error(f"Error: {exc}")
            }))
        )

        await entrypoint.run()
        ```

    Attributes:
        _fastmcp: The FastMCP application instance.
        _run_args: Positional arguments to pass to fastmcp.run_async().
        _run_kwargs: Keyword arguments to pass to fastmcp.run_async().
        _plugins: List of plugins applied to this entrypoint.

    """

    def __init__(self, fastmcp: FastMCP, *args: Any, **kwargs: Any) -> None:
        """Initialize the FastMCP entrypoint.

        Args:
            fastmcp: The FastMCP application instance.
            *args: The arguments to pass to the FastMCP run_async method.
            **kwargs: The keyword arguments to pass to the FastMCP run_async method.

        """

        self._fastmcp = fastmcp
        self._run_args = args
        self._run_kwargs = kwargs
        self._events = EventEmitter[Self]()
        self._plugins: list[AbstractEntrypointPlugin[Self]] = []
        self._plugin_registry = PluginRegistry[Self]()

    @property
    def version(self) -> str:
        """Get the version of the FastMCP entrypoint."""
        return "1.0.0"

    @property
    def events(self) -> EventEmitter[Self]:
        """Get the event emitter for the entrypoint."""
        return self._events

    def use_plugin(self, plugin: AbstractEntrypointPlugin[Self]) -> Self:
        """Add and apply a plugin to the entrypoint.

        Plugins are applied immediately and stored for lifecycle hooks.

        Args:
            plugin: The plugin to add and apply.

        Returns:
            Self for method chaining.

        """
        return apply_plugin(self, plugin, self._plugin_registry)

    def use_preset(
        self,
        preset: AbstractEntrypointPluginPreset[
            Self,
            AbstractEntrypointPlugin[Self],
        ],
    ) -> Self:
        """Add and apply a plugin preset to the entrypoint.

        Args:
            preset: The plugin preset to apply.

        Returns:
            Self for method chaining.

        """
        return apply_preset(self, preset, self._plugin_registry)

    async def startup(self) -> None:
        """Startup the FastMCP entrypoint.

        Prepares the FastMCP application for execution. Calls startup hooks for all plugins.
        This method is called before run() and should be idempotent.

        Raises:
            EntrypointInconsistencyError: If configuration is invalid.

        """
        await self.events.emit(EntrypointStartupEvent(component=self))

    async def shutdown(self) -> None:
        """Shutdown the FastMCP entrypoint.

        Cleans up resources and stops the FastMCP application. Calls shutdown hooks for all plugins.
        This method is called after run() completes or is cancelled. Should be idempotent
        and safe to call multiple times.

        """
        await self.events.emit(EntrypointShutdownEvent(component=self))

    def get_app(self) -> FastMCP:
        """Get the FastMCP application instance.

        Returns:
            The FastMCP application instance.

        """
        return self._fastmcp

    @property
    def plugin_registry(self) -> PluginRegistry:
        """Get the plugin registry for plugin discovery.

        Provides read-only access to the plugin registry, allowing plugins
        to discover other plugins without accessing private attributes.

        Returns:
            The plugin registry instance.

        """
        return self._plugin_registry

    async def run(self) -> None:
        """Run the FastMCP entrypoint.

        Starts the FastMCP application and begins processing requests.
        This method runs indefinitely until cancelled or an error occurs.

        Raises:
            Exception: Any exception that occurs during execution.

        """
        await self._fastmcp.run_async(*self._run_args, **self._run_kwargs)



"""Conftest for idempotency."""

import pytest_asyncio
from dishka import AsyncContainer
from fastapi import FastAPI
from httpx import ASGITransport, AsyncClient

from haolib.entrypoints.fastapi import FastAPIEntrypoint
from haolib.entrypoints.plugins.fastapi import FastAPIDishkaPlugin, FastAPIIdempotencyMiddlewarePlugin


@pytest_asyncio.fixture
async def app(container: AsyncContainer) -> FastAPI:
    """Test app."""
    app_instance = FastAPI()
    FastAPIEntrypoint(app=app_instance).use_plugin(FastAPIDishkaPlugin(container)).use_plugin(
        FastAPIIdempotencyMiddlewarePlugin()
    )
    return app_instance


@pytest_asyncio.fixture
async def test_client(app: FastAPI) -> AsyncClient:
    """Test client."""
    return AsyncClient(transport=ASGITransport(app), base_url="http://test")



"""Redis idempotency keys storage."""

from datetime import timedelta

from redis.asyncio import Redis

from haolib.web.idempotency.storages.abstract import AbstractIdempotencyKeysStorage


class RedisIdempotencyKeysStorage(AbstractIdempotencyKeysStorage):
    """Redis idempotency keys storage."""

    def __init__(self, redis: Redis, ttl: timedelta) -> None:
        """Initialize the redis idempotency keys storage.

        Args:
            redis: Redis instance.
            ttl: Time to live for the idempotency key.

        """
        self._redis = redis
        self._ttl = ttl

    async def is_processed(self, idempotency_key: str) -> bool:
        """Check if the idempotency key is processed.

        Args:
            idempotency_key: The idempotency key.

        Returns:
            bool: True if the idempotency key is processed, False otherwise.

        """
        response = await self._redis.get(idempotency_key)

        return response is not None

    async def set_processed(self, idempotency_key: str) -> None:
        """Set the idempotency key as processed.

        Args:
            idempotency_key: The idempotency key.

        """
        await self._redis.set(idempotency_key, idempotency_key, px=self._ttl)



"""Concrete operation classes for type-safe storage operations.

Each operation type has its own class with typed parameters.
This allows for better type safety and IDE autocomplete.
"""

from __future__ import annotations

from collections.abc import AsyncIterator, Callable, Iterable
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, TypeVar, overload

if TYPE_CHECKING:
    from haolib.pipelines.base import Operation, Pipeline, TargetBoundOperation
    from haolib.storages.indexes.abstract import SearchIndex
    from haolib.storages.targets.abstract import AbstractDataTarget
else:
    # Import SearchIndex for runtime (needed for isinstance checks)
    from haolib.storages.indexes.abstract import SearchIndex

# Import Operation for runtime (needed for class inheritance)
# Must be imported after TYPE_CHECKING block to avoid circular import
from haolib.pipelines.base import Operation, Pipeline, TargetBoundOperation

T_Data = TypeVar("T_Data")


@dataclass(frozen=True)
class CreateOperation[T_Data](Operation[Iterable[T_Data], list[T_Data]]):
    """Create data in storage.

    Example:
        ```python
        users = [User(id=1, name="Alice"), User(id=2, name="Bob")]
        await storage.execute(CreateOperation(data=users))
        ```

    """

    data: list[T_Data]
    """Data to create."""

    @overload  # type: ignore[override]
    def __or__(
        self,
        other: FilterOperation[T_Data],
    ) -> Pipeline[Iterable[T_Data], list[T_Data], list[T_Data]]: ...

    @overload  # type: ignore[override]
    def __or__[T_Result](
        self,
        other: MapOperation[T_Data, T_Result],
    ) -> Pipeline[Iterable[T_Data], list[T_Data], list[T_Result]]: ...

    @overload  # type: ignore[override]
    def __or__[T_Result](
        self,
        other: ReduceOperation[T_Data, T_Result],
    ) -> Pipeline[Iterable[T_Data], list[T_Data], T_Result]: ...

    @overload  # type: ignore[override]
    def __or__[T_Result](
        self,
        other: TransformOperation[list[T_Data], T_Result],
    ) -> Pipeline[Iterable[T_Data], list[T_Data], T_Result]: ...

    @overload  # type: ignore[override]
    def __or__(
        self,
        other: TargetBoundOperation[list[T_Data]],
    ) -> Pipeline[Iterable[T_Data], list[T_Data], list[T_Data]]: ...

    def __or__[T_NextResult](
        self,
        other: Operation[Any, T_NextResult] | TargetBoundOperation[Any],
    ) -> Pipeline[Iterable[T_Data], list[T_Data], T_NextResult]:
        """Compose with next operation, preserving type information."""
        from haolib.pipelines.base import Pipeline  # noqa: PLC0415

        return Pipeline(first=self, second=other)


@dataclass(frozen=True)
class ReadOperation[T_Data](Operation[Any, AsyncIterator[T_Data]]):
    """Read data from storage with type-safe index.

    Example:
        ```python
        from haolib.storages.indexes import index

        user_index = index(User, age=25)
        async for user in await storage.execute(ReadOperation(search_index=user_index)):
            print(user)
        ```

    """

    search_index: SearchIndex[T_Data]
    """Typed search index. Must contain data_type."""

    @overload  # type: ignore[override]
    def __or__(
        self,
        other: FilterOperation[T_Data],
    ) -> Pipeline[Any, AsyncIterator[T_Data], list[T_Data]]: ...

    @overload  # type: ignore[override]
    def __or__[T_Result](
        self,
        other: MapOperation[T_Data, T_Result],
    ) -> Pipeline[Any, AsyncIterator[T_Data], list[T_Result]]: ...

    @overload  # type: ignore[override]
    def __or__[T_Result](
        self,
        other: ReduceOperation[T_Data, T_Result],
    ) -> Pipeline[Any, AsyncIterator[T_Data], T_Result]: ...

    @overload  # type: ignore[override]
    def __or__[T_Result](
        self,
        other: TransformOperation[AsyncIterator[T_Data], T_Result],
    ) -> Pipeline[Any, AsyncIterator[T_Data], T_Result]: ...

    @overload  # type: ignore[override]
    def __or__(
        self,
        other: CreateOperation[T_Data],
    ) -> Pipeline[Any, AsyncIterator[T_Data], list[T_Data]]: ...

    @overload  # type: ignore[override]
    def __or__(
        self,
        other: PatchOperation[T_Data],
    ) -> Pipeline[Any, AsyncIterator[T_Data], list[T_Data]]: ...

    @overload  # type: ignore[override]
    def __or__(
        self,
        other: UpdateOperation[T_Data],
    ) -> Pipeline[Any, AsyncIterator[T_Data], list[T_Data]]: ...

    @overload  # type: ignore[override]
    def __or__(
        self,
        other: TargetBoundOperation[AsyncIterator[T_Data]],
    ) -> Pipeline[Any, AsyncIterator[T_Data], AsyncIterator[T_Data]]: ...

    def __or__[T_NextResult](
        self,
        other: Operation[Any, T_NextResult] | TargetBoundOperation[Any],
    ) -> Pipeline[Any, AsyncIterator[T_Data], T_NextResult]:
        """Compose with next operation, preserving type information."""
        from haolib.pipelines.base import Pipeline  # noqa: PLC0415

        return Pipeline(first=self, second=other)


@dataclass(frozen=True)
class PatchOperation[T_Data](Operation[Any, list[T_Data]]):
    """Partial update operation (PATCH).

    Updates only specified fields. For HTTP, uses PATCH method.
    For SQL, updates only specified columns.

    Can work in two modes:
    1. Search mode: uses search_index to find data to update
    2. Pipeline mode: uses previous_result as data to update

    Example:
        ```python
        from haolib.storages.indexes import index

        # Search mode
        user_index = index(User, id=123)
        await storage.execute(PatchOperation(
            search_index=user_index,
            patch={"is_active": True}
        ))

        # Pipeline mode (using fluent API)
        storage.read(user_index).returning() | storage.patch({"is_active": True})
        ```

    """

    search_index: SearchIndex[T_Data] | None = None
    """Index for finding data to update (if no previous_result)."""

    patch: dict[str, Any] | None = None
    """Partial update specification (if no previous_result)."""


@dataclass(frozen=True)
class UpdateOperation[T_Data](Operation[Any, list[T_Data]]):
    """Full update operation (PUT).

    Replaces entire object. For HTTP, uses PUT method.
    For SQL, updates all fields.

    Can work in two modes:
    1. Search mode: uses search_index to find data, data to update
    2. Pipeline mode: uses previous_result as data to update

    Example:
        ```python
        from haolib.storages.indexes import index

        # Search mode
        user_index = index(User, id=123)
        await storage.execute(UpdateOperation(
            search_index=user_index,
            data=User(id=123, name="John", email="john@example.com", is_active=True)
        ))

        # Pipeline mode (using fluent API)
        storage.read(user_index).returning() | storage.update(data=lambda u: User(id=u.id, name=u.name.upper()))
        ```

    """

    search_index: SearchIndex[T_Data] | None = None
    """Index for finding data to update (if no previous_result)."""

    data: T_Data | Callable[[T_Data], T_Data] | None = None
    """Full replacement data or transform function (if no previous_result)."""


@dataclass(frozen=True)
class DeleteOperation[T_Data](Operation[Any, int]):
    """Delete data from storage with type-safe index.

    Can work in two modes:
    1. Search mode: uses search_index to find data to delete
    2. Pipeline mode: uses previous_result as data to delete

    Example:
        ```python
        from haolib.storages.indexes import index

        # Search mode
        user_index = index(User, id=123)
        deleted_count = await storage.execute(DeleteOperation(search_index=user_index))

        # Pipeline mode (using fluent API)
        storage.read(user_index).returning() | storage.delete()
        ```

    """

    search_index: SearchIndex[T_Data] | None = None
    """Typed search index for finding data to delete (if no previous_result)."""


@dataclass(frozen=True)
class FilterOperation[T_Data](Operation[Iterable[T_Data], list[T_Data]]):
    """Filter data.

    Example:
        ```python
        pipeline = (
            ReadOperation(search_index=user_index)
            | FilterOperation(predicate=lambda u: u.age >= 18)
        )
        results = await storage.execute(pipeline)
        ```

    """

    predicate: Callable[[T_Data], bool]
    """Function that returns True for items to keep."""

    @overload  # type: ignore[override]
    def __or__(
        self,
        other: FilterOperation[T_Data],
    ) -> Pipeline[Iterable[T_Data], list[T_Data], list[T_Data]]: ...

    @overload  # type: ignore[override]
    def __or__[T_Result](
        self,
        other: MapOperation[T_Data, T_Result],
    ) -> Pipeline[Iterable[T_Data], list[T_Data], list[T_Result]]: ...

    @overload  # type: ignore[override]
    def __or__[T_Result](
        self,
        other: ReduceOperation[T_Data, T_Result],
    ) -> Pipeline[Iterable[T_Data], list[T_Data], T_Result]: ...

    @overload  # type: ignore[override]
    def __or__[T_Result](
        self,
        other: TransformOperation[list[T_Data], T_Result],
    ) -> Pipeline[Iterable[T_Data], list[T_Data], T_Result]: ...

    @overload  # type: ignore[override]
    def __or__(
        self,
        other: CreateOperation[T_Data],
    ) -> Pipeline[Iterable[T_Data], list[T_Data], list[T_Data]]: ...

    @overload  # type: ignore[override]
    def __or__(
        self,
        other: PatchOperation[T_Data],
    ) -> Pipeline[Iterable[T_Data], list[T_Data], list[T_Data]]: ...

    @overload  # type: ignore[override]
    def __or__(
        self,
        other: UpdateOperation[T_Data],
    ) -> Pipeline[Iterable[T_Data], list[T_Data], list[T_Data]]: ...

    @overload  # type: ignore[override]
    def __or__(
        self,
        other: TargetBoundOperation[list[T_Data]],
    ) -> Pipeline[Iterable[T_Data], list[T_Data], list[T_Data]]: ...

    def __or__[T_NextResult](
        self,
        other: Operation[Any, T_NextResult] | TargetBoundOperation[Any],
    ) -> Pipeline[Iterable[T_Data], list[T_Data], T_NextResult]:
        """Compose with next operation, preserving type information."""
        from haolib.pipelines.base import Pipeline  # noqa: PLC0415

        return Pipeline(first=self, second=other)


@dataclass(frozen=True)
class MapOperation[T_Data, T_Result](Operation[Iterable[T_Data], list[T_Result]]):
    """Apply function to each element.

    Example:
        ```python
        pipeline = (
            ReadOperation(search_index=user_index)
            | MapOperation(mapper=lambda user, idx: user.email)
        )
        emails = await storage.execute(pipeline)
        ```

    """

    mapper: Callable[[T_Data, int], T_Result]
    """Function (item, index) -> result."""

    @overload  # type: ignore[override]
    def __or__[T_NextResult](
        self,
        other: MapOperation[T_Result, T_NextResult],
    ) -> Pipeline[Iterable[T_Data], list[T_Result], list[T_NextResult]]: ...

    @overload  # type: ignore[override]
    def __or__(
        self,
        other: FilterOperation[T_Result],
    ) -> Pipeline[Iterable[T_Data], list[T_Result], list[T_Result]]: ...

    @overload  # type: ignore[override]
    def __or__[T_ReduceResult](
        self,
        other: ReduceOperation[T_Result, T_ReduceResult],
    ) -> Pipeline[Iterable[T_Data], list[T_Result], T_ReduceResult]: ...

    @overload  # type: ignore[override]
    def __or__(
        self,
        other: ReduceOperation[Any, Any],
    ) -> Pipeline[Iterable[T_Data], list[T_Result], Any]: ...

    @overload  # type: ignore[override]
    def __or__[T_TransformResult](
        self,
        other: TransformOperation[list[T_Result], T_TransformResult],
    ) -> Pipeline[Iterable[T_Data], list[T_Result], T_TransformResult]: ...

    @overload  # type: ignore[override]
    def __or__(
        self,
        other: CreateOperation[T_Result],
    ) -> Pipeline[Iterable[T_Data], list[T_Result], list[T_Result]]: ...

    @overload  # type: ignore[override]
    def __or__(
        self,
        other: PatchOperation[T_Result],
    ) -> Pipeline[Iterable[T_Data], list[T_Result], list[T_Result]]: ...

    @overload  # type: ignore[override]
    def __or__(
        self,
        other: UpdateOperation[T_Result],
    ) -> Pipeline[Iterable[T_Data], list[T_Result], list[T_Result]]: ...

    @overload  # type: ignore[override]
    def __or__(
        self,
        other: TargetBoundOperation[list[T_Result]],
    ) -> Pipeline[Iterable[T_Data], list[T_Result], list[T_Result]]: ...

    def __or__[T_NextResult](
        self,
        other: Operation[Any, T_NextResult] | TargetBoundOperation[Any],
    ) -> Pipeline[Iterable[T_Data], list[T_Result], T_NextResult]:
        """Compose with next operation, preserving type information."""
        from haolib.pipelines.base import Pipeline  # noqa: PLC0415

        return Pipeline(first=self, second=other)


@dataclass(frozen=True)
class ReduceOperation[T_Data, T_Result](Operation[Iterable[T_Data], T_Result]):
    """Reduce collection to a single value.

    Example:
        ```python
        pipeline = (
            ReadOperation(search_index=user_index)
            | MapOperation(mapper=lambda u, idx: u.age)
            | ReduceOperation(reducer=lambda acc, age: acc + age, initial=0)
        )
        total_age = await storage.execute(pipeline)
        ```

    """

    reducer: Callable[[T_Result, T_Data], T_Result]
    """Function (accumulator, item) -> accumulator."""

    initial: T_Result
    """Initial value for accumulator."""

    @overload  # type: ignore[override]
    def __or__[T_TransformResult](
        self,
        other: TransformOperation[T_Result, T_TransformResult],
    ) -> Pipeline[Iterable[T_Data], T_Result, T_TransformResult]: ...

    @overload  # type: ignore[override]
    def __or__(
        self,
        other: FilterOperation[T_Result],
    ) -> Pipeline[Iterable[T_Data], T_Result, list[T_Result]]: ...

    @overload  # type: ignore[override]
    def __or__[T_MapResult](
        self,
        other: MapOperation[T_Result, T_MapResult],
    ) -> Pipeline[Iterable[T_Data], T_Result, list[T_MapResult]]: ...

    @overload  # type: ignore[override]
    def __or__(
        self,
        other: CreateOperation[T_Result],
    ) -> Pipeline[Iterable[T_Data], T_Result, list[T_Result]]: ...

    @overload  # type: ignore[override]
    def __or__(
        self,
        other: TargetBoundOperation[T_Result],
    ) -> Pipeline[Iterable[T_Data], T_Result, T_Result]: ...

    def __or__[T_NextResult](
        self,
        other: Operation[Any, T_NextResult] | TargetBoundOperation[Any],
    ) -> Pipeline[Iterable[T_Data], T_Result, T_NextResult]:
        """Compose with next operation, preserving type information."""
        from haolib.pipelines.base import Pipeline  # noqa: PLC0415

        return Pipeline(first=self, second=other)


@dataclass(frozen=True)
class TransformOperation[T_Data, T_Result](Operation[T_Data, T_Result]):
    """Transform data.

    Example:
        ```python
        pipeline = (
            ReadOperation(search_index=user_index)
            | TransformOperation(transformer=lambda users: [u.model_dump() for u in users])
        )
        result = await storage.execute(pipeline)
        ```

    """

    transformer: Callable[[T_Data], T_Result]
    """Function for transformation."""

    @overload  # type: ignore[override]
    def __or__[T_NextResult](
        self,
        other: TransformOperation[T_Result, T_NextResult],
    ) -> Pipeline[T_Data, T_Result, T_NextResult]: ...

    @overload  # type: ignore[override]
    def __or__(
        self,
        other: FilterOperation[T_Result],
    ) -> Pipeline[T_Data, T_Result, list[T_Result]]: ...

    @overload  # type: ignore[override]
    def __or__[T_MapResult](
        self,
        other: MapOperation[T_Result, T_MapResult],
    ) -> Pipeline[T_Data, T_Result, list[T_MapResult]]: ...

    @overload  # type: ignore[override]
    def __or__[T_ReduceResult](
        self,
        other: ReduceOperation[T_Result, T_ReduceResult],
    ) -> Pipeline[T_Data, T_Result, T_ReduceResult]: ...

    @overload  # type: ignore[override]
    def __or__(
        self,
        other: CreateOperation[T_Result],
    ) -> Pipeline[T_Data, T_Result, list[T_Result]]: ...

    @overload  # type: ignore[override]
    def __or__(
        self,
        other: TargetBoundOperation[T_Result],
    ) -> Pipeline[T_Data, T_Result, T_Result]: ...

    def __or__[T_NextResult](
        self,
        other: Operation[Any, T_NextResult] | TargetBoundOperation[Any],
    ) -> Pipeline[T_Data, T_Result, T_NextResult]:
        """Compose with next operation, preserving type information."""
        from haolib.pipelines.base import Pipeline  # noqa: PLC0415

        return Pipeline(first=self, second=other)



"""Integration tests for SQLAlchemy storage operations.

Tests basic CRUD operations and pipelines for both SQLAlchemyOperationsHandler
and SQLAlchemyStorage.
"""

from typing import Any

import pytest

from haolib.pipelines import filtero, mapo, reduceo, transformo
from haolib.pipelines.base import Pipeline
from haolib.pipelines.operations import (
    CreateOperation,
    DeleteOperation,
    FilterOperation,
    MapOperation,
    PatchOperation,
    ReadOperation,
    ReduceOperation,
    TransformOperation,
    UpdateOperation,
)
from haolib.storages.indexes.params import ParamIndex
from haolib.storages.operations.sqlalchemy import SQLAlchemyOperationsHandler
from tests.integration.storages.conftest import User, UserModel

# Constants for test values
MIN_USERS_COUNT = 2
SINGLE_USER_COUNT = 1
ALICE_AGE = 25
BOB_AGE = 30
CHARLIE_AGE = 35
DAVID_AGE = 40
EVE_AGE = 28
FRANK_AGE = 32
GRACE_AGE = 27
GRACE_UPDATED_AGE = 28
HENRY_AGE = 29
HENRY_UPDATED_AGE = 30
IVAN_AGE = 31
JACK_AGE = 25
FILTER_MIN_AGE = 30
MARY_AGE = 29
NICK_AGE = 33
RACHEL_AGE = 26
STEVE_AGE = 29
EXPECTED_TOTAL_AGE = 87  # 26 + 29 + 32
KATE_AGE = 28
LEO_AGE = 31
EXPECTED_TOTAL_AGE_2 = 59  # 28 + 31


class TestOperationsHandler:
    """Integration tests for SQLAlchemyOperationsHandler."""

    @pytest.mark.asyncio
    async def test_execute_create(self, sqlalchemy_storage: Any, registry: Any) -> None:
        """Test execute_create operation."""
        handler = SQLAlchemyOperationsHandler(
            registry=registry,
            relationship_load_depth=2,
            storage=sqlalchemy_storage,
        )

        users = [
            User(name="Alice", age=ALICE_AGE, email="alice@example.com"),
            User(name="Bob", age=BOB_AGE, email="bob@example.com"),
        ]

        operation = CreateOperation(data=users)
        # Create transaction for handler method
        txn = sqlalchemy_storage._begin_transaction()
        async with txn:
            result = await handler.execute_create(operation, txn)

        assert len(result) == MIN_USERS_COUNT
        assert result[0].name == "Alice"
        assert result[0].age == ALICE_AGE
        assert result[0].email == "alice@example.com"
        assert result[0].id is not None
        assert result[1].name == "Bob"
        assert result[1].age == BOB_AGE
        assert result[1].email == "bob@example.com"
        assert result[1].id is not None

    @pytest.mark.asyncio
    async def test_execute_create_without_registration(self, sqlalchemy_storage: Any, registry: Any) -> None:
        """Test execute_create with unregistered types (assumes storage model)."""
        handler = SQLAlchemyOperationsHandler(
            registry=registry,
            relationship_load_depth=2,
            storage=sqlalchemy_storage,
        )

        # Create storage models directly
        # Note: Even though we pass UserModel, it gets converted back to User
        # because registry has mapping for UserModel -> User
        user_models = [
            UserModel(name="Charlie", age=CHARLIE_AGE, email="charlie@example.com"),
            UserModel(name="David", age=DAVID_AGE, email="david@example.com"),
        ]

        operation = CreateOperation(data=user_models)
        # Create transaction for handler method
        txn = sqlalchemy_storage._begin_transaction()
        async with txn:
            result = await handler.execute_create(operation, txn)

        assert len(result) == MIN_USERS_COUNT
        # Result is converted back to User through registry
        assert isinstance(result[0], User)
        assert result[0].name == "Charlie"
        assert result[0].id is not None

    @pytest.mark.asyncio
    async def test_execute_read(self, sqlalchemy_storage: Any, registry: Any) -> None:
        """Test execute_read operation."""
        handler = SQLAlchemyOperationsHandler(
            registry=registry,
            relationship_load_depth=2,
            storage=sqlalchemy_storage,
        )

        # Create users first
        users = [
            User(name="Eve", age=EVE_AGE, email="eve@example.com"),
            User(name="Frank", age=FRANK_AGE, email="frank@example.com"),
        ]
        # Create transaction for handler methods
        txn = sqlalchemy_storage._begin_transaction()
        async with txn:
            create_op = CreateOperation(data=users)
            await handler.execute_create(create_op, txn)

        # Read all users
        index = ParamIndex(data_type=User)
        read_op = ReadOperation(search_index=index)

        # Create new transaction for read
        txn2 = sqlalchemy_storage._begin_transaction()
        async with txn2:
            results = [user async for user in handler.execute_read(read_op, txn2)]

        assert len(results) >= MIN_USERS_COUNT
        # Check that we got our users
        emails = {user.email for user in results}
        assert "eve@example.com" in emails
        assert "frank@example.com" in emails

    @pytest.mark.asyncio
    async def test_execute_update_with_dict_patch(self, sqlalchemy_storage: Any, registry: Any) -> None:
        """Test execute_update with dict patch."""
        handler = SQLAlchemyOperationsHandler(
            registry=registry,
            relationship_load_depth=2,
            storage=sqlalchemy_storage,
        )

        # Create user
        user = User(name="Grace", age=GRACE_AGE, email="grace@example.com")
        create_op = CreateOperation(data=[user])
        # Create transaction for handler methods
        txn = sqlalchemy_storage._begin_transaction()
        async with txn:
            created = await handler.execute_create(create_op, txn)
        user_id = created[0].id

        # Update user (partial update)
        index = ParamIndex(data_type=User, id=user_id)
        patch_op = PatchOperation(search_index=index, patch={"age": GRACE_UPDATED_AGE, "name": "Grace Updated"})

        txn = sqlalchemy_storage._begin_transaction()
        async with txn:
            updated = await handler.execute_patch(patch_op, txn)

        assert len(updated) == SINGLE_USER_COUNT
        assert updated[0].id == user_id
        assert updated[0].name == "Grace Updated"
        assert updated[0].age == GRACE_UPDATED_AGE
        assert updated[0].email == "grace@example.com"

    @pytest.mark.asyncio
    async def test_execute_update_with_callable_data(self, sqlalchemy_storage: Any, registry: Any) -> None:
        """Test execute_update with callable data."""
        handler = SQLAlchemyOperationsHandler(
            registry=registry,
            relationship_load_depth=2,
            storage=sqlalchemy_storage,
        )

        # Create user
        user = User(name="Henry", age=HENRY_AGE, email="henry@example.com")
        create_op = CreateOperation(data=[user])
        txn = sqlalchemy_storage._begin_transaction()
        async with txn:
            created = await handler.execute_create(create_op, txn)
        user_id = created[0].id

        # Update user with callable (full update)
        def update_user(u: User) -> User:
            return User(id=u.id, name="Henry Updated", age=u.age + 1, email=u.email)

        index = ParamIndex(data_type=User, id=user_id)
        update_op = UpdateOperation(search_index=index, data=update_user)

        txn = sqlalchemy_storage._begin_transaction()
        async with txn:
            updated = await handler.execute_update(update_op, txn)

        assert len(updated) == SINGLE_USER_COUNT
        assert updated[0].id == user_id
        assert updated[0].name == "Henry Updated"
        assert updated[0].age == HENRY_UPDATED_AGE

    @pytest.mark.asyncio
    async def test_execute_delete(self, sqlalchemy_storage: Any, registry: Any) -> None:
        """Test execute_delete operation."""
        handler = SQLAlchemyOperationsHandler(
            registry=registry,
            relationship_load_depth=2,
            storage=sqlalchemy_storage,
        )

        # Create user
        user = User(name="Ivan", age=IVAN_AGE, email="ivan@example.com")
        create_op = CreateOperation(data=[user])
        txn = sqlalchemy_storage._begin_transaction()
        async with txn:
            created = await handler.execute_create(create_op, txn)
        user_id = created[0].id

        # Delete user
        index = ParamIndex(data_type=User, id=user_id)
        delete_op = DeleteOperation(search_index=index)

        txn2 = sqlalchemy_storage._begin_transaction()
        async with txn2:
            deleted_count = await handler.execute_delete(delete_op, txn2)

        assert deleted_count == SINGLE_USER_COUNT

        # Verify deleted
        read_index = ParamIndex(data_type=User, id=user_id)
        read_op = ReadOperation(search_index=read_index)
        txn3 = sqlalchemy_storage._begin_transaction()
        async with txn3:
            results = [user async for user in handler.execute_read(read_op, txn3)]
        assert len(results) == 0

    @pytest.mark.asyncio
    async def test_execute_filter(self, sqlalchemy_storage: Any, registry: Any) -> None:
        """Test execute_filter operation."""
        handler = SQLAlchemyOperationsHandler(
            registry=registry,
            relationship_load_depth=2,
            storage=sqlalchemy_storage,
        )

        # Create users
        users = [
            User(name="Jack", age=JACK_AGE, email="jack@example.com"),
            User(name="Kate", age=KATE_AGE, email="kate@example.com"),
            User(name="Leo", age=LEO_AGE, email="leo@example.com"),
        ]
        create_op = CreateOperation(data=users)
        txn = sqlalchemy_storage._begin_transaction()
        async with txn:
            await handler.execute_create(create_op, txn)

        # Filter users by age >= 30
        index = ParamIndex(data_type=User)
        read_op = ReadOperation(search_index=index)
        txn2 = sqlalchemy_storage._begin_transaction()
        async with txn2:
            read_result = handler.execute_read(read_op, txn2)
            # Collect async iterator inside transaction
            users_list = [user async for user in read_result]
        filter_op: FilterOperation[User] = FilterOperation(predicate=lambda u: u.age >= FILTER_MIN_AGE)
        txn3 = sqlalchemy_storage._begin_transaction()
        async with txn3:
            filtered = await handler.execute_filter(filter_op, txn3, users_list)

        assert len(filtered) >= 1
        assert all(user.age >= FILTER_MIN_AGE for user in filtered)

    @pytest.mark.asyncio
    async def test_execute_map(self, sqlalchemy_storage: Any, registry: Any) -> None:
        """Test execute_map operation."""
        handler = SQLAlchemyOperationsHandler(
            registry=registry,
            relationship_load_depth=2,
            storage=sqlalchemy_storage,
        )

        # Create users
        users = [
            User(name="Mary", age=MARY_AGE, email="mary@example.com"),
            User(name="Nick", age=NICK_AGE, email="nick@example.com"),
        ]
        create_op = CreateOperation(data=users)
        txn = sqlalchemy_storage._begin_transaction()
        async with txn:
            await handler.execute_create(create_op, txn)

        # Map users to their names
        index = ParamIndex(data_type=User)
        read_op = ReadOperation(search_index=index)
        txn2 = sqlalchemy_storage._begin_transaction()
        async with txn2:
            read_result = handler.execute_read(read_op, txn2)
            # Collect async iterator inside transaction
            users_list = [user async for user in read_result]
        map_op: MapOperation[User, str] = MapOperation(mapper=lambda user, _idx: user.name)
        txn3 = sqlalchemy_storage._begin_transaction()
        async with txn3:
            mapped = await handler.execute_map(map_op, txn3, users_list)

        assert len(mapped) >= MIN_USERS_COUNT
        assert "Mary" in mapped
        assert "Nick" in mapped
        assert all(isinstance(name, str) for name in mapped)

    @pytest.mark.asyncio
    async def test_execute_reduce(self, sqlalchemy_storage: Any, registry: Any) -> None:
        """Test execute_reduce operation."""
        handler = SQLAlchemyOperationsHandler(
            registry=registry,
            relationship_load_depth=2,
            storage=sqlalchemy_storage,
        )

        # Create users
        users = [
            User(name="Oscar", age=RACHEL_AGE, email="oscar@example.com"),
            User(name="Paul", age=STEVE_AGE, email="paul@example.com"),
            User(name="Quinn", age=FRANK_AGE, email="quinn@example.com"),
        ]
        create_op = CreateOperation(data=users)
        txn = sqlalchemy_storage._begin_transaction()
        async with txn:
            await handler.execute_create(create_op, txn)

        # Reduce to sum of ages
        index = ParamIndex(data_type=User)
        read_op = ReadOperation(search_index=index)
        txn2 = sqlalchemy_storage._begin_transaction()
        async with txn2:
            read_result = handler.execute_read(read_op, txn2)
            # Collect async iterator inside transaction
            users_list = [user async for user in read_result]
        reduce_op: ReduceOperation[User, int] = ReduceOperation(
            reducer=lambda acc, user: acc + user.age,
            initial=0,
        )
        txn3 = sqlalchemy_storage._begin_transaction()
        async with txn3:
            total_age = await handler.execute_reduce(reduce_op, txn3, users_list)

        assert total_age >= EXPECTED_TOTAL_AGE

    @pytest.mark.asyncio
    async def test_execute_transform(self, sqlalchemy_storage: Any, registry: Any) -> None:
        """Test execute_transform operation."""
        handler = SQLAlchemyOperationsHandler(
            registry=registry,
            relationship_load_depth=2,
            storage=sqlalchemy_storage,
        )

        # Create users
        users = [
            User(name="Rachel", age=RACHEL_AGE, email="rachel@example.com"),
            User(name="Steve", age=STEVE_AGE, email="steve@example.com"),
        ]
        create_op = CreateOperation(data=users)
        txn = sqlalchemy_storage._begin_transaction()
        async with txn:
            await handler.execute_create(create_op, txn)

        # Transform to list of emails
        index = ParamIndex(data_type=User)
        read_op = ReadOperation(search_index=index)
        txn2 = sqlalchemy_storage._begin_transaction()
        async with txn2:
            read_result = handler.execute_read(read_op, txn2)
            # Collect async iterator inside transaction
            users_list = [user async for user in read_result]

        transform_op: TransformOperation[list[User], list[str]] = TransformOperation(
            transformer=lambda users: [u.email for u in users]
        )
        txn3 = sqlalchemy_storage._begin_transaction()
        async with txn3:
            emails = await handler.execute_transform(transform_op, txn3, users_list)

        assert len(emails) >= MIN_USERS_COUNT
        assert "rachel@example.com" in emails
        assert "steve@example.com" in emails


class TestStorageOperations:
    """Integration tests for SQLAlchemyStorage operations."""

    @pytest.mark.asyncio
    async def test_execute_create_operation(self, sqlalchemy_storage: Any) -> None:
        """Test executing create operation."""
        users = [
            User(name="Alice", age=ALICE_AGE, email="alice@example.com"),
            User(name="Bob", age=BOB_AGE, email="bob@example.com"),
        ]

        result = await sqlalchemy_storage.create(users).returning().execute()

        assert len(result) == MIN_USERS_COUNT
        assert result[0].name == "Alice"
        assert result[0].id is not None
        assert result[1].name == "Bob"
        assert result[1].id is not None

    @pytest.mark.asyncio
    async def test_execute_read_operation(self, sqlalchemy_storage: Any) -> None:
        """Test executing read operation."""
        # Create users first
        users = [
            User(name="Charlie", age=KATE_AGE, email="charlie@example.com"),
            User(name="David", age=FRANK_AGE, email="david@example.com"),
        ]
        await sqlalchemy_storage.create(users).returning().execute()

        # Read all users
        index = ParamIndex(data_type=User)
        results = await sqlalchemy_storage.read(index).returning().execute()

        assert len(results) >= MIN_USERS_COUNT
        emails = {user.email for user in results}
        assert "charlie@example.com" in emails
        assert "david@example.com" in emails

    @pytest.mark.asyncio
    async def test_execute_update_operation(self, sqlalchemy_storage: Any) -> None:
        """Test executing update operation."""
        # Create user
        user = User(name="Eve", age=EVE_AGE, email="eve@example.com")
        created = await sqlalchemy_storage.create([user]).returning().execute()
        user_id = created[0].id

        # Update user (partial update)
        index = ParamIndex(data_type=User, id=user_id)
        updated = (
            await sqlalchemy_storage.read(index)
            .patch({"age": GRACE_UPDATED_AGE, "name": "Eve Updated"})
            .returning()
            .execute()
        )

        assert len(updated) == SINGLE_USER_COUNT
        assert updated[0].name == "Eve Updated"
        assert updated[0].age == GRACE_UPDATED_AGE

    @pytest.mark.asyncio
    async def test_execute_delete_operation(self, sqlalchemy_storage: Any) -> None:
        """Test executing delete operation."""
        # Create user
        user = User(name="Frank", age=IVAN_AGE, email="frank@example.com")
        created = await sqlalchemy_storage.create([user]).returning().execute()
        user_id = created[0].id

        # Delete user
        index = ParamIndex(data_type=User, id=user_id)
        deleted_count = await sqlalchemy_storage.read(index).delete().execute()

        assert deleted_count == SINGLE_USER_COUNT

    @pytest.mark.asyncio
    async def test_execute_pipeline(self, sqlalchemy_storage: Any) -> None:
        """Test executing pipeline of operations."""
        # Create users
        users = [
            User(name="Grace", age=25, email="grace@example.com"),
            User(name="Henry", age=30, email="henry@example.com"),
            User(name="Iris", age=35, email="iris@example.com"),
        ]
        await sqlalchemy_storage.create(users).returning().execute()

        # Pipeline: read -> filter -> map (all in single transaction)
        index = ParamIndex(data_type=User)
        pipeline: Pipeline[Any, Any, Any] = (
            sqlalchemy_storage.read(index).returning()
            | filtero(lambda u: u.age >= FILTER_MIN_AGE)
            | mapo(lambda u, _idx: u.name)
        )

        result = await pipeline.execute()

        assert len(result) >= MIN_USERS_COUNT
        assert "Henry" in result
        assert "Iris" in result

    @pytest.mark.asyncio
    async def test_execute_with_auto_transaction(self, sqlalchemy_storage: Any) -> None:
        """Test executing operation with automatic transaction."""
        users = [
            User(name="Jack", age=JACK_AGE, email="jack@example.com"),
        ]

        # Execute without providing transaction - should create one automatically
        result = await sqlalchemy_storage.create(users).returning().execute()

        assert len(result) == SINGLE_USER_COUNT
        assert result[0].name == "Jack"
        assert result[0].id is not None

    @pytest.mark.asyncio
    async def test_execute_reduce_operation(self, sqlalchemy_storage: Any) -> None:
        """Test executing reduce operation."""
        # Create users
        users = [
            User(name="Kate", age=KATE_AGE, email="kate@example.com"),
            User(name="Leo", age=LEO_AGE, email="leo@example.com"),
        ]
        await sqlalchemy_storage.create(users).returning().execute()

        # Reduce to sum of ages
        index = ParamIndex(data_type=User)
        pipeline = sqlalchemy_storage.read(index).returning() | reduceo(
            reducer=lambda acc, user: acc + user.age,
            initial=0,
        )

        total_age = await pipeline.execute()

        assert total_age >= EXPECTED_TOTAL_AGE_2

    @pytest.mark.asyncio
    async def test_execute_transform_operation(self, sqlalchemy_storage: Any) -> None:
        """Test executing transform operation."""
        # Create users
        users = [
            User(name="Mary", age=MARY_AGE, email="mary@example.com"),
            User(name="Nick", age=NICK_AGE, email="nick@example.com"),
        ]
        await sqlalchemy_storage.create(users).returning().execute()

        # Transform to list of emails
        # Pipeline: read -> transform (transform needs previous result from pipeline)
        index = ParamIndex(data_type=User)
        pipeline = sqlalchemy_storage.read(index).returning() | transformo(
            transformer=lambda users: [u.email for u in users]
        )
        emails = await pipeline.execute()

        assert len(emails) >= MIN_USERS_COUNT
        assert "mary@example.com" in emails
        assert "nick@example.com" in emails

    @pytest.mark.asyncio
    async def test_execute_read_update_pipeline(self, sqlalchemy_storage: Any) -> None:
        """Test executing read | update pipeline (pipeline mode)."""
        # Create user
        users = [
            User(name="PipelineUser", age=25, email="pipeline@example.com"),
        ]
        created = await sqlalchemy_storage.create(users).returning().execute()
        user_id = created[0].id

        # Read and update in pipeline
        index = ParamIndex(data_type=User, id=user_id)
        pipeline = sqlalchemy_storage.read(index).returning() | sqlalchemy_storage.update(
            data=lambda u: User(id=u.id, name=u.name.upper(), age=u.age + 1, email=u.email)
        )

        updated = await pipeline.execute()

        assert len(updated) == 1
        assert updated[0].name == "PIPELINEUSER"
        assert updated[0].age == 26

    @pytest.mark.asyncio
    async def test_execute_read_patch_pipeline(self, sqlalchemy_storage: Any) -> None:
        """Test executing read | patch pipeline (pipeline mode)."""
        # Create user
        users = [
            User(name="PatchUser", age=25, email="patch@example.com"),
        ]
        created = await sqlalchemy_storage.create(users).returning().execute()
        user_id = created[0].id

        # Read and patch in pipeline
        index = ParamIndex(data_type=User, id=user_id)
        pipeline = sqlalchemy_storage.read(index).returning() | sqlalchemy_storage.patch(patch={"age": 30})

        updated = await pipeline.execute()

        assert len(updated) == 1
        assert updated[0].name == "PatchUser"
        assert updated[0].age == 30

    @pytest.mark.asyncio
    async def test_execute_read_delete_pipeline(self, sqlalchemy_storage: Any) -> None:
        """Test executing read | delete pipeline (pipeline mode)."""
        # Create user
        users = [
            User(name="DeleteUser", age=25, email="delete@example.com"),
        ]
        created = await sqlalchemy_storage.create(users).returning().execute()
        user_id = created[0].id

        # Read and delete in pipeline
        index = ParamIndex(data_type=User, id=user_id)
        pipeline = sqlalchemy_storage.read(index).returning() | sqlalchemy_storage.delete()

        deleted_count = await pipeline.execute()

        assert deleted_count == 1

        # Verify user is deleted - read should return empty iterator
        read_index = ParamIndex(data_type=User, id=user_id)
        read_result = await sqlalchemy_storage.read(read_index).returning().execute()
        # reado returns AsyncIterator, collect it to verify it's empty
        if hasattr(read_result, "__aiter__"):
            users_list = [user async for user in read_result]
            assert len(users_list) == 0
        else:
            # If it's already a list, just check it's empty
            assert len(read_result) == 0

    @pytest.mark.asyncio
    async def test_execute_read_filter_update_pipeline(self, sqlalchemy_storage: Any) -> None:
        """Test executing read | filter | update pipeline."""
        # Create users
        users = [
            User(name="FilterUpdate1", age=25, email="filter1@example.com"),
            User(name="FilterUpdate2", age=35, email="filter2@example.com"),
        ]
        await sqlalchemy_storage.create(users).returning().execute()

        # Read, filter by age >= 30, and update
        index = ParamIndex(data_type=User)
        pipeline = (
            sqlalchemy_storage.read(index).returning()
            | filtero(lambda u: u.age >= 30)
            | sqlalchemy_storage.update(data=lambda u: User(id=u.id, name=u.name.upper(), age=u.age, email=u.email))
        )

        updated = await pipeline.execute()

        assert len(updated) == 1
        assert updated[0].name == "FILTERUPDATE2"
        assert updated[0].age == 35

    @pytest.mark.asyncio
    async def test_execute_read_map_update_pipeline(self, sqlalchemy_storage: Any) -> None:
        """Test executing read | map | update pipeline."""
        # Create user
        users = [
            User(name="MapUpdate", age=25, email="map@example.com"),
        ]
        created = await sqlalchemy_storage.create(users).returning().execute()
        user_id = created[0].id

        # Read, map to extract user, and update
        index = ParamIndex(data_type=User, id=user_id)
        pipeline = (
            sqlalchemy_storage.read(index).returning()
            | mapo(lambda u, idx: u)
            | sqlalchemy_storage.update(
                data=lambda u: User(id=u.id, name=u.name + "_updated", age=u.age, email=u.email)
            )
        )

        updated = await pipeline.execute()

        assert len(updated) == 1
        assert updated[0].name == "MapUpdate_updated"



"""Parameter-based search index."""

from typing import Any, TypeVar

from haolib.storages.indexes.abstract import SearchIndex

T_Data = TypeVar("T_Data")


class ParamIndex[T_Data](SearchIndex[T_Data]):
    """Index built from parameters.

    Used when you need to create an index from parameters
    rather than using a predefined one.

    Example:
        ```python
        # Create index from parameters
        index = ParamIndex(User, age=25, is_active=True)
        await storage.execute(reado(search_index=index))

        # Or using convenience function
        from haolib.storages.indexes import index
        user_index = index(User, age=25, is_active=True)
        await storage.execute(reado(search_index=user_index))
        ```

    """

    def __init__(
        self,
        data_type: type[T_Data],
        **params: Any,
    ) -> None:
        """Create index from parameters.

        Args:
            data_type: Type of data to search.
            **params: Search parameters (e.g., age=25, email="john@example.com").

        """
        self.__haolib_data_type__ = data_type
        self.params = params

    @property
    def data_type(self) -> type[T_Data]:
        """Type of data to search.

        Returns:
            Type of data.

        """
        return self.__haolib_data_type__


def create_index[T_Data](
    data_type: type[T_Data],
    **params: Any,
) -> ParamIndex[T_Data]:
    """Create a parameter-based index.

    Convenience function for creating ParamIndex.

    Args:
        data_type: Type of data to search.
        **params: Search parameters.

    Returns:
        ParamIndex for the given data type.

    Example:
        ```python
        from haolib.storages.indexes import create_index

        user_index = create_index(User, age=25, is_active=True)
        await storage.execute(reado(search_index=user_index))
        ```

    """
    return ParamIndex(data_type=data_type, **params)



"""SQLAlchemy transactions."""

from types import TracebackType
from typing import Self

from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.ext.asyncio import AsyncSession

from haolib.database.transactions.abstract import AbstractDatabaseTransaction, DatabaseTransactionError


class SQLAlchemyTransaction(AbstractDatabaseTransaction[AsyncSession]):
    """SQLAlchemy transaction."""

    def __init__(self, session: AsyncSession) -> None:
        """Initialize the transaction.

        Args:
            session (AsyncSession): The session to use for the transaction.

        """
        self._session: AsyncSession | None = session

    async def __aenter__(self) -> Self:
        """Start the transaction."""
        return self

    async def __aexit__(
        self, exc_type: type[BaseException] | None, exc_value: BaseException | None, traceback: TracebackType | None
    ) -> None:
        """End the transaction."""
        # We need to set the session to None to fix the problem with the garbage collection.
        # Problem: when we pass an AsyncSession instance to a repository, it is not closed even after we exit
        # the context manager of the the session. To fix it, we add another abstraction layer on top of the session
        # and set the session to None when we exit the context manager.
        if self._session is None:
            raise DatabaseTransactionError("Transaction session is closed.")

        try:
            await self._session.commit()
        except SQLAlchemyError:
            await self._session.rollback()
            raise
        finally:
            await self.close()

    async def close(self) -> None:
        """Close the session."""
        if self._session is None:
            raise DatabaseTransactionError("Transaction is already closed.")

        self._session = None

    async def get_manipulator(self) -> AsyncSession:
        """Get the session.

        Returns:
            AsyncSession: The session.

        Raises:
            DatabaseTransactionError: If the transaction is closed.

        """
        if self._session is None:
            raise DatabaseTransactionError("Transaction is closed. Did you exit the transaction context manager?")

        return self._session



"""Observability utilities."""

import sentry_sdk
from opentelemetry import trace


async def observe_exception(exc: Exception) -> None:
    """Observe exception."""

    span = trace.get_current_span()
    span.record_exception(exc)
    span.set_status(trace.Status(trace.StatusCode.ERROR, str(exc)))
    sentry_sdk.capture_exception(exc)



"""Conftest."""

import pytest_asyncio
from dishka import AsyncContainer, Provider, Scope, make_async_container, provide
from fastapi import FastAPI, Request
from fastapi.testclient import TestClient
from fastmcp import FastMCP
from httpx import ASGITransport, AsyncClient
from pydantic import Field
from redis.asyncio import Redis

from haolib.configs.base import BaseConfig
from haolib.configs.idempotency import IdempotencyConfig
from haolib.configs.observability import ObservabilityConfig
from haolib.configs.redis import RedisConfig
from haolib.configs.s3 import S3Config
from haolib.configs.sqlalchemy import SQLAlchemyConfig
from haolib.dependencies.dishka.redis import RedisProvider
from haolib.dependencies.dishka.sqlalchemy import SQLAlchemyProvider
from haolib.entrypoints.fastapi import FastAPIEntrypoint
from haolib.entrypoints.fastmcp import FastMCPEntrypoint
from haolib.entrypoints.plugins.fastapi.dishka import FastAPIDishkaPlugin
from haolib.entrypoints.plugins.fastapi.exceptions import FastAPIExceptionHandlersPlugin
from haolib.entrypoints.plugins.fastapi.fastmcp import FastAPIFastMCPPlugin
from haolib.entrypoints.plugins.fastapi.idempotency import FastAPIIdempotencyMiddlewarePlugin
from haolib.exceptions.base.fastapi import FastAPIBaseException
from haolib.exceptions.handlers.fastapi import (
    fastapi_base_exception_handler,
    fastapi_unknown_exception_handler,
    fastapi_unknown_exception_handler_with_observability,
)
from haolib.web.idempotency.storages.abstract import AbstractIdempotencyKeysStorage
from haolib.web.idempotency.storages.redis import RedisIdempotencyKeysStorage


class MockAppConfig(BaseConfig):
    """Mock app config."""

    s3: S3Config
    sqlalchemy: SQLAlchemyConfig
    redis: RedisConfig
    idempotency: IdempotencyConfig
    observability: ObservabilityConfig = Field(default_factory=ObservabilityConfig)


class MockProvider(Provider):
    """Mock provider."""

    @provide(scope=Scope.APP)
    async def app_config(self) -> MockAppConfig:
        """App config."""
        return MockAppConfig.from_env()

    @provide(scope=Scope.APP)
    async def sqlalchemy_config(self, app_config: MockAppConfig) -> SQLAlchemyConfig:
        """SQLAlchemy config."""
        return app_config.sqlalchemy

    @provide(scope=Scope.REQUEST)
    async def idempotency_keys_storage(
        self, redis: Redis, idempotency_config: IdempotencyConfig
    ) -> AbstractIdempotencyKeysStorage:
        """Idempotency keys storage."""
        return RedisIdempotencyKeysStorage(redis=redis, ttl=idempotency_config.ttl)

    @provide(scope=Scope.APP)
    async def redis_config(self, app_config: MockAppConfig) -> RedisConfig:
        """Redis config."""
        return app_config.redis

    @provide(scope=Scope.APP)
    async def idempotency_config(self, app_config: MockAppConfig) -> IdempotencyConfig:
        """Idempotency config."""
        return app_config.idempotency


@pytest_asyncio.fixture
async def container() -> AsyncContainer:
    """Test container."""
    return make_async_container(SQLAlchemyProvider(), RedisProvider(), MockProvider())


@pytest_asyncio.fixture
async def app_config(container: AsyncContainer) -> MockAppConfig:
    """App config."""
    return await container.get(MockAppConfig)


@pytest_asyncio.fixture()
async def clean_redis(container: AsyncContainer) -> None:
    """Clean Redis for testing."""
    async with container(scope=Scope.REQUEST) as nested_container:
        await (await nested_container.get(Redis)).flushdb()


@pytest_asyncio.fixture
async def app(container: AsyncContainer) -> FastAPI:
    """Test app."""
    app_instance = FastAPI()
    FastAPIEntrypoint(app=app_instance).use_plugin(FastAPIDishkaPlugin(container)).use_plugin(
        FastAPIExceptionHandlersPlugin(
            exception_handlers={
                Exception: fastapi_unknown_exception_handler,
                FastAPIBaseException: fastapi_base_exception_handler,
            }
        )
    ).use_plugin(FastAPIIdempotencyMiddlewarePlugin())
    return app_instance


@pytest_asyncio.fixture
async def app_with_observability(container: AsyncContainer) -> FastAPI:
    """Test app."""
    app_instance = FastAPI()
    FastAPIEntrypoint(app=app_instance).use_plugin(FastAPIDishkaPlugin(container)).use_plugin(
        FastAPIExceptionHandlersPlugin(
            exception_handlers={
                Exception: fastapi_unknown_exception_handler_with_observability,
                FastAPIBaseException: fastapi_base_exception_handler,
            }
        )
    ).use_plugin(FastAPIIdempotencyMiddlewarePlugin())
    return app_instance


@pytest_asyncio.fixture
async def app_with_mcp_and_mcp(container: AsyncContainer) -> tuple[FastAPI, FastMCP]:
    """Test app."""

    app = FastAPI()
    fastapi_entrypoint = (
        FastAPIEntrypoint(app=FastAPI())
        .use_plugin(FastAPIDishkaPlugin(container))
        .use_plugin(
            FastAPIExceptionHandlersPlugin(
                exception_handlers={
                    Exception: fastapi_unknown_exception_handler,
                    FastAPIBaseException: fastapi_base_exception_handler,
                }
            )
        )
        .use_plugin(FastAPIIdempotencyMiddlewarePlugin())
    )

    @app.post("/hello")
    async def hello(body: Request) -> str:
        return "hello"

    fastmcp_entrypoint = FastMCPEntrypoint(fastmcp=FastMCP.from_fastapi(app=app))
    fastapi_entrypoint.use_plugin(FastAPIFastMCPPlugin(fastmcp_entrypoint.get_app(), "/mcp"))

    return app, fastmcp_entrypoint.get_app()


@pytest_asyncio.fixture
async def test_client(app: FastAPI) -> AsyncClient:
    """Test client."""
    return AsyncClient(transport=ASGITransport(app), base_url="http://test")


@pytest_asyncio.fixture
async def test_client_without_raise_server_exceptions(app: FastAPI) -> TestClient:
    """Test client with raise server exceptions."""
    return TestClient(app, raise_server_exceptions=False)


@pytest_asyncio.fixture
async def test_client_without_raise_server_exceptions_with_observability(app_with_observability: FastAPI) -> TestClient:
    """Test client with raise server exceptions and observability."""
    return TestClient(app_with_observability, raise_server_exceptions=False)



"""Models for entities."""

from typing import Any, Protocol, Self


class AbstractCreatableFrom[T_From](Protocol):
    """Abstract model for creatable from mapped."""

    @classmethod
    def create_from(cls, from_value: T_From, *args: Any, **kwargs: Any) -> Self:
        """Create a model instance from a from value."""
        ...


class AbstractConvertable[T_To](Protocol):
    """Abstract model for convertable."""

    def convert(self, *args: Any, **kwargs: Any) -> T_To:
        """Convert the model."""
        ...


class AbstractUpdateableFrom[T_From](Protocol):
    """Abstract model for updatable from mapped."""

    def update_from(self, from_value: T_From, *args: Any, **kwargs: Any) -> Self:
        """Update the model from a from value."""
        ...


class AbstractMappedModel[T_MappedTo](AbstractCreatableFrom[T_MappedTo], AbstractConvertable[T_MappedTo], Protocol):
    """Abstract mapped model."""


class AbstractUpdateableMappedModel[T_MappedTo](
    AbstractMappedModel[T_MappedTo], AbstractUpdateableFrom[T_MappedTo], Protocol
):
    """Abstract updateable mapped model."""



"""Base operations for data pipelines.

Supports executable pipelines across multiple targets (storages, ML models, APIs, etc.)
with target binding syntax. The ^ operator semantically means "send data to this target".

Example:
        ```python
        from haolib.pipelines import reduceo, transformo
        from haolib.storages.indexes.params import ParamIndex

        # Fluent API: storage methods return composites that can be chained
        pipeline = (
            sql_storage.read(ParamIndex(User)).returning()
            | reduceo(lambda acc, u: acc + u.age, 0)
            | transformo(lambda total: str(total).encode())  # Executes in Python
            | s3_storage.create()
        )

        # Future: ML model target
        # pipeline = (
        #     sql_storage.read(...).returning()
        #     | transformo(...)  # Prepare features
        #     | ml_model.predict()  # Send to ML model
        #     | s3_storage.create()
        # )

        # Execute pipeline directly
        result = await pipeline.execute()
        ```

"""

from __future__ import annotations

from abc import ABC
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Union, overload

if TYPE_CHECKING:
    from haolib.pipelines.operations import (
        CreateOperation,
        FilterOperation,
        MapOperation,
        ReduceOperation,
        TransformOperation,
    )
    from haolib.storages.targets.abstract import AbstractDataTarget
else:
    from typing import Protocol

    class AbstractDataTarget(Protocol):  # noqa: D101
        async def execute[T_Result](  # noqa: D102
            self,
            operation: Operation[Any, T_Result] | Pipeline[Any, Any, T_Result],
        ) -> T_Result: ...


# Type alias for operations that can be composed
OperationLike = Union["Operation[Any, Any]", "Pipeline[Any, Any, Any]", "TargetBoundOperation[Any]"]


@dataclass(frozen=True)
class Operation[T_Data, T_Result](ABC):
    """Base operation class.

    Operations don't know how to execute themselves. Storage does.
    Operations are immutable and composable via | operator.

    Each concrete operation should inherit from this class and define
    its own typed parameters.

    Example:
        ```python
        from haolib.pipelines.operations import ReadOperation, FilterOperation

        op1 = ReadOperation(search_index=user_index)
        op2 = FilterOperation(predicate=lambda u: u.age >= 18)
        pipeline = op1 | op2
        ```

    """

    def __or__[T_NextResult](
        self,
        other: Operation[Any, T_NextResult] | TargetBoundOperation[Any],
    ) -> Pipeline[Any, T_Result, T_NextResult]:
        """Compose operations via | operator.

        Args:
            other: Next operation in the pipeline.

        Returns:
            Pipeline combining both operations.

        Example:
            ```python
            from haolib.pipelines import filtero
            pipeline = sql_storage.read(...).returning() | filtero(lambda u: u.age >= 18)
            ```

        """
        return Pipeline(
            first=self,
            second=other,
        )

    def __xor__(self, target: AbstractDataTarget) -> TargetBoundOperation[T_Result]:
        """Send operation to data target.

        Semantically means "send data to this target" (storage, ML model, API, etc.).

        Args:
            target: Target where operation should be executed.
                   Can be storage, ML model, API, or any other data target.

        Returns:
            TargetBoundOperation that knows where to execute.

        Example:
            ```python
            from haolib.storages.indexes.params import ParamIndex

            # Storage target (using fluent API)
            bound_op = sql_storage.read(ParamIndex(User)).returning()

            # ML Model target (future)
            # bound_op = ml_model.predict(...)
            ```

        """
        return TargetBoundOperation(operation=self, target=target)


@dataclass(frozen=True)
class Pipeline[T_Data, T_FirstResult, T_SecondResult]:
    """Pipeline of operations.

    Represents a composition of operations that will be executed sequentially.
    Results from previous operations can be passed to next operations if they support it.

    Example:
        ```python
        from haolib.pipelines.operations import ReadOperation, FilterOperation

        pipeline = (
            ReadOperation(search_index=user_index)
            | FilterOperation(predicate=lambda u: u.age >= 18)
        )
        ```

    """

    first: (
        Operation[T_Data, T_FirstResult]
        | Pipeline[Any, Any, T_FirstResult]
        | TargetBoundOperation[T_FirstResult]
        | TargetSwitch[Any, T_FirstResult]
    )
    """First operation in the pipeline."""

    second: (
        Operation[Any, T_SecondResult]
        | Pipeline[Any, Any, T_SecondResult]
        | TargetBoundOperation[T_SecondResult]
        | TargetSwitch[Any, T_SecondResult]
    )
    """Second operation in the pipeline."""

    def __or__(
        self,
        other: Operation[Any, Any] | Pipeline[Any, Any, Any] | TargetBoundOperation[Any] | Any,
    ) -> Pipeline[Any, Any, Any]:
        """Continue the pipeline.

        Args:
            other: Next operation to add to the pipeline.

        Returns:
            Extended pipeline.

        """
        # Handle composites from fluent API
        from haolib.storages.fluent.composites import BaseComposite  # noqa: PLC0415

        if isinstance(other, BaseComposite):
            # Extract operations from composite and build pipeline
            other_ops = other._operations
            if len(other_ops) == 1:
                # Single operation - bind to storage
                second_op: TargetBoundOperation[Any] = TargetBoundOperation(
                    operation=other_ops[0], target=other._storage
                )
            else:
                # Multiple operations - build pipeline and bind to storage
                other_pipeline: Pipeline[Any, Any, Any] | Operation[Any, Any] = other_ops[0]
                for op in other_ops[1:]:
                    other_pipeline = Pipeline(first=other_pipeline, second=op)
                second_op = TargetBoundOperation(operation=other_pipeline, target=other._storage)
            return Pipeline(first=self, second=second_op)

        # Check if operation needs previous_result (Python operations)
        if isinstance(other, Operation):
            from haolib.pipelines.operations import (  # noqa: PLC0415
                DeleteOperation,
                FilterOperation,
                MapOperation,
                PatchOperation,
                ReduceOperation,
                TransformOperation,
                UpdateOperation,
            )

            # Python operations (Filter, Map, Reduce, Transform) should not be bound to storage
            if isinstance(other, (FilterOperation, MapOperation, ReduceOperation, TransformOperation)):
                return Pipeline(first=self, second=other)  # type: ignore[arg-type]

        return Pipeline(
            first=self,
            second=other,
        )

    def __xor__(self, target: AbstractDataTarget) -> TargetBoundOperation[T_SecondResult]:
        """Send pipeline result to data target.

        Semantically means "send pipeline result to this target".

        Args:
            target: Target where pipeline result should be executed.

        Returns:
            TargetBoundOperation for the pipeline result.

        Example:
            ```python
            from haolib.pipelines import filtero
            pipeline = (
                sql_storage.read(user_index).returning()
                | filtero(lambda u: u.age >= 18)
            )
            ```

        """
        return TargetBoundOperation(operation=self, target=target)

    def validate(self) -> None:
        """Validate pipeline structure before execution.

        Raises:
            PipelineValidationError: If pipeline structure is invalid.

        Example:
            ```python
            from haolib.pipelines import filtero
            pipeline = (
                sql_storage.read(...).returning()
                | filtero(lambda u: u.age >= 18)
                | s3_storage.create()
            )

            # Validate before execution
            pipeline.validate()
            result = await pipeline.execute()
            ```

        """
        from haolib.pipelines.validator import PipelineValidator  # noqa: PLC0415

        validator = PipelineValidator()
        validator.validate(self)

    async def execute(self) -> T_SecondResult:
        """Execute pipeline across multiple targets.

        Automatically validates pipeline before execution.
        Detects which target to use for each operation,
        optimizes operations on target side when possible, and coordinates
        data transfer between targets.

        Targets can be storages, ML models, APIs, or any other data processing target.

        Returns:
            Final pipeline result.

        Raises:
            PipelineValidationError: If pipeline structure is invalid.
            ValueError: If no target is specified for an operation that requires it.
            RuntimeError: If pipeline execution fails.

        Example:
            ```python
            from haolib.pipelines import reduceo, transformo
            from haolib.storages.indexes.params import ParamIndex

            # Fluent API: storage methods return composites
            pipeline = (
                sql_storage.read(ParamIndex(User)).returning()
                | reduceo(lambda acc, u: acc + u.age, 0)
                | transformo(lambda total: str(total).encode())  # Executes in Python
                | s3_storage.create()
            )

            # Execute pipeline directly (validation happens automatically)
            result = await pipeline.execute()
            ```

        """
        # Validate before execution
        self.validate()

        from haolib.pipelines.executor import ExecutablePipelineExecutor  # noqa: PLC0415

        executor = ExecutablePipelineExecutor()
        return await executor.execute(self)


@dataclass(frozen=True)
class TargetBoundOperation[T_Result]:
    """Operation bound to a specific data target.

    Created when using ^ operator to send operation to a target.
    Operations without target binding execute in Python.

    Targets can be storages, ML models, APIs, or any other data processing target.
    The ^ operator semantically means "send data to this target".

    Example:
            ```python
            from haolib.storages.indexes.params import ParamIndex

            # Storage target (using fluent API)
            bound_op = sql_storage.read(ParamIndex(User)).returning()

            # ML Model target (future)
            # bound_op = ml_model.predict(...)
            ```

    """

    operation: Operation[Any, T_Result] | Pipeline[Any, Any, T_Result]
    """Operation to execute."""
    target: AbstractDataTarget
    """Target where operation should be executed."""

    async def execute(self) -> T_Result:
        """Execute the bound operation in the target.

        Allows direct execution of bound operations without needing to call
        storage.execute() explicitly.

        Returns:
            Result of operation execution.

        Raises:
            TargetError: If target operation fails.
            TypeError: If operation type is not supported.

        Example:
            ```python
            from haolib.storages.indexes.params import ParamIndex

            # Read operation (using fluent API)
            users = await sql_storage.read(ParamIndex(User)).returning().execute()

            # Create operation
            await sql_storage.create([user1, user2]).returning().execute()

            # Update operation
            await sql_storage.read(ParamIndex(User, id=1)).update(User(...)).returning().execute()

            # Delete operation
            await sql_storage.read(ParamIndex(User, id=1)).delete().execute()
            ```

        """
        return await self.target.execute(self.operation)

    def __or__[T_NextResult](
        self,
        other: Operation[Any, T_NextResult] | TargetBoundOperation[Any],
    ) -> Pipeline[Any, T_Result, T_NextResult]:
        """Continue pipeline after target-bound operation.

        Args:
            other: Next operation in the pipeline.

        Returns:
            Pipeline combining operations.

        """
        # If next operation is also bound to target, check if we need to switch
        if isinstance(other, TargetBoundOperation):
            if other.target != self.target:
                # Target switch needed
                # self.operation can be Operation, Pipeline, or TargetSwitch
                # TargetSwitch.source_result expects Operation or Pipeline, not TargetSwitch
                # So we need to unwrap if it's a TargetSwitch
                source_op = self.operation
                if isinstance(source_op, TargetSwitch):
                    # If source is already a TargetSwitch, use its source_result
                    source_op = source_op.source_result

                return Pipeline(
                    first=self,
                    second=TargetSwitch(
                        source_result=source_op,
                        source_target=self.target,
                        target_target=other.target,
                        next_operation=other.operation,
                    ),
                )
            # Same target - continue normally
            return Pipeline(first=self, second=other)

        # Next operation is not bound to target
        # Check if it needs previous_result (Python operations)
        from haolib.pipelines.operations import (  # noqa: PLC0415
            DeleteOperation,
            FilterOperation,
            MapOperation,
            PatchOperation,
            ReduceOperation,
            TransformOperation,
            UpdateOperation,
        )

        if isinstance(other, (FilterOperation, MapOperation, ReduceOperation, TransformOperation)):
            # Python operations - don't bind to target
            return Pipeline(first=self, second=other)  # type: ignore[arg-type]

        # Storage operations - bind to same target
        bound_other = TargetBoundOperation(operation=other, target=self.target)
        return Pipeline(first=self, second=bound_other)

    def __xor__(self, target: AbstractDataTarget) -> TargetBoundOperation[T_Result]:
        """Rebind to different target.

        Args:
            target: New target to bind to.

        Returns:
            New TargetBoundOperation with different target.

        """
        return TargetBoundOperation(operation=self.operation, target=target)


@dataclass(frozen=True)
class TargetSwitch[T_Data, T_Result]:
    """Operation that switches execution context to another target.

    Created automatically when pipeline switches between targets.
    Handles data transfer and context switching.

    Targets can be storages, ML models, APIs, or any other data processing target.

    """

    source_result: Operation[Any, T_Data] | Pipeline[Any, Any, T_Data]
    """Source operation that produced the data."""
    source_target: AbstractDataTarget
    """Target where source operation was executed."""
    target_target: AbstractDataTarget
    """Target where next operation will be executed."""
    next_operation: Operation[Any, T_Result] | Pipeline[Any, Any, T_Result]
    """Next operation to execute in target target."""

    def __or__[T_NextResult](
        self,
        other: Operation[Any, T_NextResult] | TargetBoundOperation[Any],
    ) -> Pipeline[Any, T_Result, T_NextResult]:
        """Continue pipeline after target switch.

        Args:
            other: Next operation in the pipeline.

        Returns:
            Extended pipeline.

        """
        return Pipeline(first=self, second=other)

    def __xor__(self, target: AbstractDataTarget) -> TargetBoundOperation[T_Result]:
        """Bind target switch result to another target.

        Args:
            target: Target to bind to.

        Returns:
            TargetBoundOperation with next_operation bound to target.

        Note:
            TargetSwitch is typically created automatically during pipeline composition.
            Binding a TargetSwitch to a target binds its next_operation to that target.
            TargetSwitch itself should not be wrapped in TargetBoundOperation.

        """
        # Bind the next_operation to the target, not the switch itself
        # TargetSwitch should not be wrapped in TargetBoundOperation
        if isinstance(self.next_operation, TargetBoundOperation):
            # Already bound, just return it
            return self.next_operation
        return TargetBoundOperation(operation=self.next_operation, target=target)



"""Health check executor with timeout and parallel execution support."""

import asyncio
import time
from collections.abc import Sequence
from datetime import timedelta
from typing import Any

from haolib.web.health.checkers.abstract import (
    AbstractHealthChecker,
    HealthCheckMetadata,
    HealthCheckResult,
    HealthStatus,
)

# Constants
TIMEOUT_ERROR_MESSAGE = "Health check timed out"


class HealthCheckExecutor:
    """Executes health checks with timeout and parallel execution support.

    Example:
        ```python
        executor = HealthCheckExecutor(timeout=timedelta(seconds=5))
        results = await executor.execute([db_checker, redis_checker])
        overall_status = executor.aggregate_status(results)
        ```

    """

    def __init__(
        self,
        timeout: timedelta | None = None,
        execute_parallel: bool = True,
    ) -> None:
        """Initialize the health check executor.

        Args:
            timeout: Maximum time to wait for all checks to complete.
                If None, no timeout is applied.
            execute_parallel: Whether to execute checks in parallel.
                If False, checks are executed sequentially.

        """
        self.timeout = timeout
        self.execute_parallel = execute_parallel

    async def execute(
        self,
        checkers: Sequence[AbstractHealthChecker],
    ) -> list[HealthCheckResult]:
        """Execute health checks.

        Args:
            checkers: Sequence of health checkers to execute.
                Must not be empty.

        Returns:
            List of health check results in the same order as checkers.
            Returns empty list if checkers is empty.

        Raises:
            ValueError: If checkers is None (not empty sequence).

        """
        if not checkers:
            return []

        if self.execute_parallel:
            return await self._execute_parallel(checkers)
        return await self._execute_sequential(checkers)

    async def _execute_parallel(
        self,
        checkers: Sequence[AbstractHealthChecker],
    ) -> list[HealthCheckResult]:
        """Execute checks in parallel."""
        tasks = [asyncio.create_task(self._execute_with_timing(checker)) for checker in checkers]

        if self.timeout is not None:
            try:
                results = await asyncio.wait_for(
                    asyncio.gather(*tasks, return_exceptions=True),
                    timeout=self.timeout.total_seconds(),
                )
            except TimeoutError:
                # Cancel remaining tasks and create timeout results
                for task in tasks:
                    if not task.done():
                        task.cancel()
                # Wait for cancellations to complete
                await asyncio.gather(*tasks, return_exceptions=True)
                results = [
                    HealthCheckResult(
                        metadata=self._get_checker_metadata(checker, i),
                        is_healthy=False,
                        error=TIMEOUT_ERROR_MESSAGE,
                    )
                    for i, checker in enumerate(checkers)
                ]
        else:
            results = await asyncio.gather(*tasks, return_exceptions=True)

        # Convert exceptions to unhealthy results
        final_results: list[HealthCheckResult] = []
        for i, result in enumerate(results):
            if isinstance(result, Exception):
                checker = checkers[i]
                final_results.append(
                    HealthCheckResult(
                        metadata=self._get_checker_metadata(checker, i),
                        is_healthy=False,
                        error=str(result),
                    )
                )
            elif isinstance(result, HealthCheckResult):
                final_results.append(result)
            else:
                # Fallback for unexpected types
                checker = checkers[i]
                final_results.append(
                    HealthCheckResult(
                        metadata=self._get_checker_metadata(checker, i),
                        is_healthy=False,
                        error=f"Unexpected result type: {type(result)}",
                    )
                )

        return final_results

    async def _execute_sequential(
        self,
        checkers: Sequence[AbstractHealthChecker],
    ) -> list[HealthCheckResult]:
        """Execute checks sequentially."""
        results: list[HealthCheckResult] = []
        for i, checker in enumerate(checkers):
            try:
                if self.timeout is not None:
                    result = await asyncio.wait_for(
                        self._execute_with_timing(checker),
                        timeout=self.timeout.total_seconds(),
                    )
                    results.append(result)
                else:
                    result = await self._execute_with_timing(checker)
                    results.append(result)
            except TimeoutError:
                results.append(
                    HealthCheckResult(
                        metadata=self._get_checker_metadata(checker, i),
                        is_healthy=False,
                        error=TIMEOUT_ERROR_MESSAGE,
                    )
                )
            except Exception as e:
                results.append(
                    HealthCheckResult(
                        metadata=self._get_checker_metadata(checker, i),
                        is_healthy=False,
                        error=str(e),
                    )
                )
        return results

    async def _execute_with_timing(self, checker: AbstractHealthChecker) -> HealthCheckResult:
        """Execute a checker and measure its duration.

        Args:
            checker: The health checker to execute.

        Returns:
            HealthCheckResult: The result with measured duration.

        """
        start_time = time.perf_counter()
        result = await checker()
        duration_ms = (time.perf_counter() - start_time) * 1000

        # Update duration if not already set by the checker
        if result.duration_ms is None:
            result.duration_ms = duration_ms

        return result

    @staticmethod
    def _get_checker_metadata(checker: Any, index: int) -> HealthCheckMetadata:
        """Extract metadata from checker or create default.

        Args:
            checker: The health checker instance.
            index: The index of the checker in the sequence.

        Returns:
            HealthCheckMetadata: Metadata for the checker.

        """
        if hasattr(checker, "metadata") and isinstance(checker.metadata, HealthCheckMetadata):
            return checker.metadata

        if hasattr(checker, "name") and isinstance(checker.name, str):
            return HealthCheckMetadata(name=checker.name, critical=True)

        return HealthCheckMetadata(name=f"checker_{index}", critical=True)

    @staticmethod
    def aggregate_status(results: list[HealthCheckResult]) -> HealthStatus:
        """Aggregate health check results into overall status.

        Rules:
        - If any critical check is unhealthy, status is UNHEALTHY
        - If any non-critical check is unhealthy, status is DEGRADED
        - Otherwise, status is HEALTHY

        Args:
            results: List of health check results.

        Returns:
            Overall health status.

        """
        if not results:
            return HealthStatus.HEALTHY

        has_unhealthy_critical = any(not result.is_healthy and result.metadata.critical for result in results)
        if has_unhealthy_critical:
            return HealthStatus.UNHEALTHY

        has_unhealthy_non_critical = any(not result.is_healthy and not result.metadata.critical for result in results)
        if has_unhealthy_non_critical:
            return HealthStatus.DEGRADED

        return HealthStatus.HEALTHY



"""SQL query index for relational databases."""

from typing import Any, TypeVar

from sqlalchemy import Delete, Select, Update

from haolib.storages.indexes.abstract import SearchIndex

T_Data = TypeVar("T_Data")

# Type alias for SQLAlchemy statements
SQLStatement = Select[Any] | Update | Delete


class SQLQueryIndex[T_Data](SearchIndex[T_Data]):
    """SQL index for relational databases (SQLAlchemy).

    Used for SQL-based storages like PostgreSQL, MySQL, etc.
    The query can be a SQLAlchemy Select, Update, or Delete statement.
    Data type is automatically extracted from the query when needed.

    Example:
        ```python
        from sqlalchemy import select, update

        # For ReadOperation (using fluent API)
        index = SQLQueryIndex(
            query=select(UserModel).where(UserModel.email == "john@example.com")
        )
        await storage.read(index).returning().execute()

        # For UpdateOperation (optimized, using fluent API)
        update_index = SQLQueryIndex(
            query=update(UserModel).where(UserModel.id == 1).values(name="John")
        )
        await storage.read(update_index).update(User(...)).returning().execute()
        ```

    """

    def __init__(
        self,
        query: SQLStatement,
    ) -> None:
        """Create SQL query index.

        Args:
            query: SQLAlchemy query (Select, Update, or Delete).

        """
        self.__haolib_query__ = query

    @property
    def query(self) -> SQLStatement:
        """SQLAlchemy query (Select, Update, or Delete).

        Returns:
            SQLAlchemy query.

        """
        return self.__haolib_query__

    @property
    def data_type(self) -> type[T_Data]:
        """Type of data to search.

        Note: This property requires a DataTypeRegistry to extract the type from query.
        In practice, the type is extracted automatically by IndexHandler when needed.

        Returns:
            Type of data (extracted from query via registry).

        Raises:
            NotImplementedError: If called directly without registry context.
                In practice, IndexHandler._get_data_type_from_query() should be used instead.

        """
        # This property is required by SearchIndex protocol, but in practice
        # the type is extracted by IndexHandler._get_data_type_from_query().
        # We raise NotImplementedError to indicate this should not be called directly.
        msg = (
            "SQLQueryIndex.data_type cannot be accessed directly. "
            "Type is automatically extracted from query by IndexHandler when needed."
        )
        raise NotImplementedError(msg)

    def __repr__(self) -> str:
        """String representation."""
        return f"SQLQueryIndex(query={self.query!r})"

    def __eq__(self, other: object) -> bool:
        """Equality comparison."""
        if not isinstance(other, SQLQueryIndex):
            return False
        return self.query == other.query

    def __hash__(self) -> int:
        """Hash for frozen dataclass-like behavior."""
        return hash(id(self.query))



"""Plugin registry for component plugins."""

from collections import defaultdict
from collections.abc import Iterator
from typing import TYPE_CHECKING, cast

from haolib.components.abstract import ComponentInconsistencyError
from haolib.components.plugins.versioning import check_version_compatibility

if TYPE_CHECKING:
    from haolib.components.abstract import AbstractComponent
    from haolib.components.plugins.abstract import AbstractPlugin


class PluginRegistry[T_Component: AbstractComponent]:
    """Registry for component plugins.

    Provides type-safe methods for plugins to discover other plugins.

    Example:
        ```python
        # In a plugin's apply() method
        if entrypoint.plugin_registry.has_plugin(FastAPIDishkaPlugin):
            # FastAPIDishkaPlugin is available
            fastapi_dishka_plugin = entrypoint.plugin_registry.get_plugin(FastAPIDishkaPlugin)
            if fastapi_dishka_plugin is not None:
                # Access plugin's public methods
                container = fastapi_dishka_plugin.get_container()
        ```

    """

    def __init__(self) -> None:
        """Initialize the plugin registry."""
        self._plugins: list[AbstractPlugin[T_Component]] = []

    def add(self, new_plugin: AbstractPlugin[T_Component], component_version: str | None = None) -> None:
        """Add a plugin to the registry.

        This method should only be called by the component during plugin application.
        Uses topological sort to order plugins based on dependencies.

        Raises:
            ComponentInconsistencyError: If circular dependencies detected.

        Args:
            new_plugin: The plugin to register.
            component_version: The component version to check for compatibility.

        Note:
            This is an internal method. Plugins are automatically registered when
            using component.use_plugin() or component.use_preset().

        """
        # Check version compatibility if component version is provided
        if component_version is not None:
            metadata = new_plugin.metadata
            check_version_compatibility(
                component_version=component_version,
                plugin_name=metadata.name,
                min_version=metadata.min_component_version,
                max_version=metadata.max_component_version,
            )

        self._plugins.append(new_plugin)

        # Build dependency graph
        plugin_map: dict[type[AbstractPlugin], AbstractPlugin] = {
            type(new_plugin): new_plugin for new_plugin in self._plugins
        }

        # Add already registered plugins to graph
        # Build adjacency list
        graph: dict[type[AbstractPlugin], set[type[AbstractPlugin]]] = defaultdict(set)
        in_degree: dict[type[AbstractPlugin], int] = defaultdict(int)

        for plugin in self._plugins:
            plugin_type = type(plugin)
            in_degree[plugin_type] = 0

            for dep_type in plugin.dependencies:
                if dep_type not in plugin_map:
                    msg = f"{plugin_type.__name__} requires {dep_type.__name__} which is not available"
                    raise ComponentInconsistencyError(msg)
                graph[dep_type].add(plugin_type)
                in_degree[plugin_type] += 1

        # Topological sort
        queue: list[type[AbstractPlugin]] = [plugin_type for plugin_type, degree in in_degree.items() if degree == 0]
        result: list[AbstractPlugin[T_Component]] = []

        while queue:
            # Sort by priority for deterministic ordering
            queue.sort(key=lambda plugin_type: plugin_map[plugin_type].priority)
            current = queue.pop(0)
            result.append(plugin_map[current])

            for dependent in graph[current]:
                in_degree[dependent] -= 1
                if in_degree[dependent] == 0:
                    queue.append(dependent)

        # Check for circular dependencies
        if len(result) != len(self._plugins):
            raise ComponentInconsistencyError("Circular dependency detected in plugin dependencies")

        self._plugins = result

    def has_plugin[T: AbstractPlugin](self, plugin_type: type[T]) -> bool:
        """Check if a plugin of the given type is registered.

        Args:
            plugin_type: The plugin class to check for.

        Returns:
            True if a plugin of the given type is registered.

        Example:
            ```python
            if registry.has_plugin(FastAPIDishkaPlugin):
                # FastAPIDishkaPlugin is available
            ```

        """
        return any(isinstance(plugin, plugin_type) for plugin in self._plugins)

    def get_plugin[T: AbstractPlugin](self, plugin_type: type[T]) -> T | None:
        """Get a plugin of the given type.

        Returns the first plugin instance that matches the given type.
        If multiple instances of the same type exist, returns the first one added.

        Args:
            plugin_type: The plugin class to retrieve.

        Returns:
            The plugin instance if found, None otherwise.

        Example:
            ```python
            fastapi_dishka_plugin = registry.get_plugin(FastAPIDishkaPlugin)
            if fastapi_dishka_plugin is not None:
                # Use the plugin instance or access plugin's public methods
                container = fastapi_dishka_plugin.get_container()
            ```

        """
        for plugin in self._plugins:
            if isinstance(plugin, plugin_type):
                return cast("T", plugin)
        return None

    def get_all_plugins(self) -> Iterator[AbstractPlugin[T_Component]]:
        """Get all registered plugins.

        Returns an iterator over all registered plugins in the order they were added.
        Each call creates a new iterator, so multiple iterations are independent.

        Returns:
            An iterator over all registered plugins.

        Example:
            ```python
            for plugin in registry.get_all_plugins():
                # Process each plugin
                print(type(plugin).__name__)
            ```

        """
        return iter(self._plugins)



"""Redis config."""

from pydantic import Field, RedisDsn
from pydantic_settings import BaseSettings


class RedisConfig(BaseSettings):
    """Redis config.

    This config is used to configure the redis.

    Attributes:
        url (RedisDsn): The url of the redis. Defaults to "redis://localhost:6379".

    """

    url: RedisDsn = Field(default=RedisDsn("redis://localhost:6379"))



"""Index handler for SQLAlchemy storage.

Converts various index types to SQLAlchemy queries.
"""

from typing import Any

from sqlalchemy import Delete, Select, Update, select
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import DeclarativeBase

from haolib.storages.data_types.registry import DataTypeRegistry
from haolib.storages.indexes.abstract import SearchIndex
from haolib.storages.indexes.params import ParamIndex
from haolib.storages.indexes.sql import SQLQueryIndex


class IndexHandler:
    """Handler for converting indexes to SQLAlchemy queries."""

    def __init__(self, registry: DataTypeRegistry) -> None:
        """Initialize the index handler.

        Args:
            registry: Data type registry for converting user types to storage models.

        """
        self._registry = registry

    def _extract_model_from_query(self, query: Select[Any] | Update[Any] | Delete[Any]) -> type[DeclarativeBase] | None:
        """Extract storage model from SQLAlchemy query (Select, Update, or Delete).

        Args:
            query: SQLAlchemy query (Select, Update, or Delete).

        Returns:
            Storage model type, or None if cannot be extracted.

        """
        # For Update and Delete statements, get table from statement
        if isinstance(query, (Update, Delete)):
            # Update and Delete have .table attribute
            if hasattr(query, "table"):
                table = query.table
                # Check if it's a mapped class
                if isinstance(table, type) and issubclass(table, DeclarativeBase):
                    return table
                # Check if it's a Table and get mapped class from it
                if (
                    hasattr(table, "entity")
                    and isinstance(table.entity, type)
                    and issubclass(table.entity, DeclarativeBase)
                ):
                    return table.entity

        # For Select queries, use existing logic
        if isinstance(query, Select):
            # Try to get model from query.get_final_froms() (tables/models in FROM clause)
            froms = query.get_final_froms() if hasattr(query, "get_final_froms") else getattr(query, "froms", None)
            if froms:
                for from_item in froms:
                    # Check if it's a mapped class (DeclarativeBase)
                    if isinstance(from_item, type) and issubclass(from_item, DeclarativeBase):
                        return from_item
                    # Check if it's a Table and get mapped class from it
                    if (
                        hasattr(from_item, "entity")
                        and isinstance(from_item.entity, type)
                        and issubclass(from_item.entity, DeclarativeBase)
                    ):
                        return from_item.entity

            # Try to get from column_descriptions (for Select queries)
            if hasattr(query, "column_descriptions") and query.column_descriptions:
                for desc in query.column_descriptions:
                    if "entity" in desc and desc["entity"] is not None:
                        entity = desc["entity"]
                        if isinstance(entity, type) and issubclass(entity, DeclarativeBase):
                            return entity

            # Try to get from selected_columns
            if hasattr(query, "selected_columns"):
                for col in query.selected_columns:
                    if hasattr(col, "entity") and col.entity is not None:
                        entity = col.entity
                        if isinstance(entity, type) and issubclass(entity, DeclarativeBase):
                            return entity

        return None

    def get_data_type_from_query(self, query: Select[Any] | Update[Any] | Delete[Any]) -> type[Any]:
        """Get user data type from SQLAlchemy query.

        Extracts storage model from query and finds corresponding user type from registry.

        Args:
            query: SQLAlchemy query (Select, Update, or Delete).

        Returns:
            User data type.

        Raises:
            ValueError: If cannot extract model from query or model is not registered.

        """
        # Extract storage model from query
        storage_model = self._extract_model_from_query(query)
        if storage_model is None:
            msg = "Cannot extract model type from SQLAlchemy query. Please ensure query uses a mapped model (e.g., select(UserModel))."
            raise ValueError(msg)

        # Find user type from registry
        registration = self._registry.get_for_storage_type(storage_model)
        if not registration:
            msg = f"Storage model {storage_model} is not registered in DataTypeRegistry"
            raise ValueError(msg)

        return registration.user_type

    async def build_query[T_Data](
        self,
        index: SearchIndex[T_Data],
        session: AsyncSession,
    ) -> Select[tuple[Any]] | Update[Any] | Delete[Any]:
        """Convert index to SQLAlchemy query.

        Args:
            index: Search index to convert.
            session: SQLAlchemy session (for potential query execution).

        Returns:
            SQLAlchemy query (Select, Update, or Delete).

        Raises:
            TypeError: If index type is not supported.
            ValueError: If storage model is not registered for data type.

        """
        if isinstance(index, SQLQueryIndex):
            # Already a SQLAlchemy query
            return index.query

        if isinstance(index, ParamIndex):
            return await self._build_param_index_query(index, session)

        msg = f"Unsupported index type: {type(index)}"
        raise TypeError(msg)

    async def _build_param_index_query[T_Data](
        self,
        index: ParamIndex[T_Data],
        session: AsyncSession,  # noqa: ARG002
    ) -> Select[tuple[Any]]:
        """Build query from ParamIndex.

        Args:
            index: ParamIndex to convert.
            session: SQLAlchemy session.

        Returns:
            SQLAlchemy Select query.

        Raises:
            ValueError: If storage model is not registered for data type.

        """
        # Get storage model from registry
        registration = self._registry.get_for_user_type(index.data_type)
        if not registration:
            msg = f"No storage model registered for {index.data_type}"
            raise ValueError(msg)

        model: type[DeclarativeBase] = registration.storage_type
        query: Select[tuple[Any]] = select(model)

        # Apply parameters as filters
        for key, value in index.params.items():
            if hasattr(model, key):
                column = getattr(model, key)
                query = query.where(column == value)

        return query



"""SQLAlchemy config."""

from pydantic import Field
from pydantic_settings import BaseSettings


class SQLAlchemyConfig(BaseSettings):
    """SQLAlchemy config.

    This config is used to configure the sqlalchemy.

    Attributes:
        url (str): The url of the sqlalchemy.
        use_pool (bool): Whether to use a connection pool. Defaults to False.

    """

    url: str
    use_pool: bool = Field(default=False)



"""Filter types."""

from haolib.enums.base import BaseEnum


class FilterType(BaseEnum):
    """Filter type."""

    eq = "eq"
    """Equals."""
    ne = "ne"
    """Not equals."""
    in_list = "in_list"
    """In list."""
    not_in_list = "not_in_list"
    """Not in list."""
    gt = "gt"
    """Greater than."""
    ge = "ge"
    """Greater than or equal."""
    lt = "lt"
    """Less than."""
    le = "le"
    """Less than or equal."""
    like = "like"
    """Like."""
    ilike = "ilike"
    """Case-insensitive like."""
    order_by = "order_by"
    """Order by (ASC or DESC)."""
    skip = "skip"
    """Skip auto-generated filter."""
    func = "func"
    """Function filter."""


class MinMaxType(BaseEnum):
    """Min max type."""

    MIN = "min"
    MAX = "max"


class OrderByType(BaseEnum):
    """Order by filter type."""

    ASC = "asc"
    DESC = "desc"


class IntervalUnit(BaseEnum):
    """Interval unit."""

    SECOND = "second"
    MINUTE = "minute"
    HOUR = "hour"
    DAY = "day"
    WEEK = "week"
    MONTH = "month"
    QUARTER = "quarter"
    YEAR = "year"



"""FastAPI exceptions plugin."""

from collections.abc import Callable
from typing import Any

from haolib.entrypoints.fastapi import FastAPIEntrypoint
from haolib.entrypoints.plugins.abstract import AbstractEntrypointPlugin
from haolib.exceptions.base.fastapi import FastAPIBaseException
from haolib.exceptions.handlers.fastapi import fastapi_base_exception_handler, fastapi_unknown_exception_handler


class FastAPIExceptionHandlersPlugin(AbstractEntrypointPlugin[FastAPIEntrypoint]):
    """Plugin for adding exception handlers to FastAPI entrypoints.

    Example:
        ```python
        exception_handlers = {
            ValueError: lambda request, exc: JSONResponse(status_code=400, content={"error": str(exc)})
        }
        entrypoint = FastAPIEntrypoint(app=app).use_plugin(FastAPIExceptionHandlersPlugin(exception_handlers))
        ```

    """

    def __init__(self, exception_handlers: dict[type[Exception], Callable[..., Any]] | None = None) -> None:
        """Initialize the exception handlers plugin.

        Args:
            exception_handlers: Dictionary mapping exception types to handler functions.
                If None, uses default handlers for Exception and FastAPIBaseException.

        """
        if exception_handlers is None:
            exception_handlers = {
                Exception: fastapi_unknown_exception_handler,
                FastAPIBaseException: fastapi_base_exception_handler,
            }
        self._exception_handlers = exception_handlers

    def apply(self, component: FastAPIEntrypoint) -> FastAPIEntrypoint:
        """Apply exception handlers to the entrypoint.

        Args:
            component: The FastAPI entrypoint to configure.

        Returns:
            The configured entrypoint.

        """
        for exception, handler in self._exception_handlers.items():
            component.get_app().add_exception_handler(exception, handler)
        return component
